// ****************************************************************
//
//    Class sparse
//    
//    Description:  
//      Sparse is a class for storing, and manipulating sparse
//      matrices.  It is realy just an interface to the 
//      SPARSKIT library written by Professor Youcef Saad.
//
//    Problems:
//      Currently it only stores the values of the matrices
//      as doubles.  Later it should be templated to support
//      integers as well.
//
//      If you discover any bugs, it would be very helpful if
//      you sent me e-mail.  Thank you.
//
//      Author:                Peter Braaten
//      e-mail:                braa0007@mermaid.micro.umn.edu
//      Date last modified:    July 28, 1994
//
// ****************************************************************

/*========================================================================
				PRIVATE :
				--------

    int nrow;		// row dimension of the sparse matrix
    int ncol;		// col dimension of the sparse matrix
    int nnz;		// number of non-zero elements
    int *ia;		// array containing the pointers to the beginning of
			// each column
    int *ja;		// array containing the row indices of the elements
			// stored in a
    double *a;		// array containing the non zero elements
    char title[73];
    char key[9];
    char type[4];
				FUNCTIONS :
				----------

constructors
destructors

get_nrow:	   returns the row dimension of the matrix
get_ncol : 	   returns the col dimension of the matrix
get_nnz : 	   returns the number of non zero elements
get_element(i,j) : returns the value of the element whose coordinates are (i,j)
get_ia : 	   returns ia
get_ja : 	   returns ja
get_a : 	   returns a
set_element(i,j,v) : sets the element whose coordinates are (i,j) to the value v
readmt(f) : 	   reads the matrix stored in the Harwell-Boeing file f
prtmt(f) : 	   prints the matrix in the Harwell-Boeing file f
gen57pt :	   generates a 5-point or 7-point matrix (see genmat.f)
submat :	   extracts a submatrix from a sparse one.

region_avg :       returns the average value from a certain region of a matrix

filter : 	   filters elements from a matrix according to their magnitude
csort  : 	   sorts the elements in increasing order of columns
getdia : 	   extracts a specified diagonal from a matrix.
rmdia : 	   removes a specified diagonal from a matrix.
getu :	  	   extracts the upper triangular part of a matrix
getl :		   extracts the lower triangular part of a matrix
getbwd :	   gives the bandwith of the upper and the lower part of the mat
getbwdu :	   gives the bandwith of the upper part of the matrix
getbwdl :	   gives the bandwith of the lower part of the matrix
cnrms : 	   computes the norms of the columns
roscal : 	   scales the rows of a matrix by their norms
coscal : 	   scales the columns of a matrix by their norms
init_zero : 	   initializes all the elements of the matrix to zero
init_identity :    makes the matrix into an identity matrix
indset0 :	   greedy algorithm for independent set ordering
indset1 :	   idem  but using minimal degree traversal
multic,colors :    greedy algorithm for multicoloring (the only difference
		   between these functions is what they return)

operators:         =, <<=, plus, minus, *

filter(mat,f) :    applies the function f to all the elements of mat
output(mat) :      outputs the matrix mat
rperm :            calculates the row permutation of a sparse matrix
transp :           calculates the transpose of a sparse matrix
floor, trunc, ceiling, round, abs

FUNCTIONS ONLY USED BY SOLVE_SYSTEM (not available in spmath)

amux  :		   multilplies a matrix with a vector
atmux :		   multiplies the transpose of a matrix with a vector
ilut, iluk : 	   preconditionner
ilud, ilu0 :       preconditionner
lusol, lutsol:     preconditionner
msrcsr : 	   create a matrix in compressed row format from a matrix
		   in modified row format 
csrmsr :	   create a matrix in modified row format from a matrix 
		   in compressed row format
solve_system :	   solve a linear system Ax = y where A is a sparse matrix.


========================================================================*/

#ifndef SPARSE_H
#define SPARSE_H


#include <stdio.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <math.h>
#include <iostream.h>
#include <stdlib.h>
#include "csparse.h"


extern "C" { typedef double (*ffunc)(double); }

typedef struct {

  int    Dimension;                        // (given)
  int    NNz;                              // (given)
  double Average_NNZ_Per_Column;           // avnz_col
  double Averate_NNZ_Std_Deviation;        // avnz_col
  int    NNZ_In_Strict_Lower_Part;         // nonz_lud
  int    NNZ_In_Strict_Upper_Part;         // nonz_lud
  int    NNZ_In_Main_Diagonal;             // nonz_lud
  int    Weight_Of_Longest_Column;         // nonz 
  int    Weight_Of_Shortest_Column;        // nonz
  int    Weight_Of_Longest_Row;            // nonz
  int    Weight_Of_Shortest_Row;           // nonz
  int    Matching_Elements_In_Symmetry;    // ansym
  double Relative_Symmetry_Match;          // ansy
  double Average_Distance_From_Diag;       // distaij
  double Average_Distance_Std_Deviation;   // distaij

  double Frobenius_Norm;                   // frobnorm
  double Frobenius_Norm_Of_Symmetric;      // ansym
  double Frobenius_Norm_Of_Nonsymmetric;   // ansym
  double Maximum_Element;
  double Percentage_Diag_Dominant_Rows;    // diag_domi
  double Percentage_Diag_Dominant_Columns; // diag_domi

  int    Lower_Bandwidth;                  // bandwidth
  int    Upper_Bandwidth;                  // bandwidth
  int    Maximum_Bandwidth;                // bandwidth
  double  Average_Bandwidth;                // bandwidth
  int    NNZ_In_Skyline_Storage;           // skyline
  int    Ninety_Percent_Bandwidth;         // bandpart
  int    Eighty_Percent_Bandwidth;          // bandpart
  int    N_Nonvoid_Diagonals;
  int    N_Important_Diagonals;             // n_imp_diag
  int    Most_Important_Diagonals[10];     // n_imp_diag
  double Important_Diag_Accumulated_Percentages[10]; // n_imp_diag

  int    Has_Block_Structure;              // blkfnd
  int    Block_Size;                       // blkfnd

  int    N_Zero_Rows;                      // nonz
  int    N_Zero_Columns;                   // nonz

} sparse_stats; // end struct sparse_stats

class sparse {

  public:
    // Constructors
    sparse();
    sparse (int, int, int, double *, int *, int *);
    sparse (const sparse &);
    sparse (int, int, int);
    sparse (int, int, double *);
    sparse (int, int, int, int);

    void free();

    // Destructor
    ~sparse();

    // Selectors
    int get_nrow () const;
    int get_ncol () const;
    int get_nnz () const;
    double get_element (int, int) const;
    int* get_ia () const;  
    int* get_ja () const;  
    double *get_guess () const;
    double *get_rhs () const;
    double *get_exact () const;

    void set_guess(double *);
    void set_rhs(double *);
    void set_exact(double *);

    const char* get_title() const;
    const char* get_type() const;
    const char* get_key() const;
    double* get_a() const;

    // Functions
    void set_element (int, int, double);
    void readmt (char *);
    void gen57pt (int, int, int, double, double, double, double, double, double 	);
    void gen57pt (int, int, int, double, double, double, double, double, double 	, int *, double *);

    double max_element();
    void submat(int, int, int, int, int);
    double region_avg(int, int, int, int);
    double region_avg2(int, int, int, int, int);
    void filter (int, int, double);
    void csort (int);
    double * getdia(int);
    void rmdia(int); 
    void getl();
    void getu();
    int * getbwd();
    int getbwdl();
    int getbwdu();
    int blkfnd();
    double * cnrms(int );
    void roscal(int);
    void coscal(int);
    void addblk(sparse, int, int);
    int * get1up();
    int * infdia();
    void prtmt (const char *);
    void init_zero (int, int);
    void init_identity (int);  

    // Operators
    void operator = (const sparse &);
    void operator <<= (sparse &);

    int ilut(int, double);
    int ilut_msr(int, double);
    int ilutp(int, double, double);
    int ilutp_msr(int, double, double);
    int ilud(double, double);
    int ilud_msr(double, double);
    int iluk_msr(int);
    int ilu0_msr();
    void msrcsr(double *, int *, int);
    void csrmsr();
    void amux(double *, double *);
    void atmux(double *,double *);
    void lusol(int, double *, double *);
    void lutsol(int, double *, double *);
    double * runriters(double *, double *,char *);
    double * solve_system(double *, double *, char *, char *, double *, int, int, int, double, double);
    void multic();
    int size();
    int * colors();
    int * il_multic();
    int *iord_multic();
    void indset0();
    void indset1();

    void get_stats(sparse_stats &);

  private:
    // Data
    int nrow;		// row dimension of the sparse matrix
    int ncol;		// col dimension of the sparse matrix
    int nnz;		// number of non-zero elements
    int *ia;		// array containing the pointers to the beginning of
			// each column
    int *ja;		// array containing the row indices of the elements
			// stored in a
    double *a;		// array containing the non zero elements

    double *guess;
    double *rhs;
    double *exact;

    char title[73];
    char key[9];
    char type[4];

}; // end class sparse

void output (sparse &);
sparse filter (const sparse &, ffunc);

sparse operator + (sparse &, sparse &);
sparse operator - (sparse &, sparse &);
sparse operator - (sparse &);
sparse operator * (sparse &, sparse &);
sparse operator * (double, sparse &);
sparse operator * (sparse &, double);
sparse operator / (sparse &, double);

sparse floor (sparse &);    
sparse ceiling (sparse &);    
sparse trunc (sparse &);    
sparse round (sparse &);    
sparse abs (sparse &);    
sparse transp (sparse &);
sparse rperm (sparse &, int *);
sparse cperm (sparse &, int *);
sparse dperm (sparse &, int *, int *, int);
int irnd(int, double);
double rnd0nb(int, double);

/*  These functions are defined in csparse.c and are used in
    sparse.C.  They are wrappers for SPARSKIT routines. */

extern "C" {

  void get_dim (char *, int *, int *, int *);
  void gen57pt_matrix(int, int, int, double *, int, int *, double *,
		      int *, int *, int *, double *);
  void submat_matrix(int, int, int, int, int, int, double *, int *, 
		     int *, int *, int *, double *, int *, int *);
  void filter_matrix(int, int, double, double *, int *, int *,
		     double *, int *, int *, int , int );
  void csort_matrix (int, double *, int *, int *, int *, int);
  void getdia_matrix (int, int, int, double *, int *, int *, int *, 
		      double *, int *, int);
  void getu_matrix(int, double *, int *, int *, double *, int *, int *);
  void getl_matrix(int, double *, int *, int *, double *, int *, int *);
  void getbwd_matrix(int, double *, int *, int *, int *, int *);
  void blkfnd_matrix(int, int *, int *, int *);
  void cnrms_matrix(int, int, double *, int *, int *, double *);
  void coscal_matrix(int, int, int, double *, int *, int *, double *, 
		     double *, int *, int *, int);
  void roscal_matrix(int, int, int, double *, int *, int *, double *, 
		     double *, int *, int *, int);
  void addblk_matrices(int, int, double *, int *, int *, int, int, int, 
		       int, double *, int *, int *, int *, int *, double *, 
		       int *, int *, int,  int);
  void get1up_matrix (int, int *, int *, int *);
  void infdia_matrix (int, int *, int *, int *, int *);
  void read_matrix (char *, int , int , int , double *, 
		    int *, int *, char *, char *, char *);
  void write_matrix (char *, int , int , int , double *, 
		     int *, int *, char *, char *, char *, int);
  void multiply_matrices (int, int, double *, int *, int *,
			  double *, int *, int *, 
			  double *, int *, int *, int, int *);
  void add_matrices (int, int, double *, int *, int *,
		     double *, int *, int *, 
		     double *, int *, int *, int, int *);
  void subtract_matrices (int, int, double *, int *, int *,
			  double *, int *, int *, 
			  double *, int *, int *, int);
  void transpose_matrix (int, int, double *, int *, int *, int *);
  void rperm_matrix (int, double *, int *, int *, 
		     double *, int *, int *, int *);
  void cperm_matrix (int, double *, int *, int *, 
		     double *, int *, int *, int *);
  void dperm_matrix (int, double *, int *, int *, 
		     double *, int *, int *, int *, int *, int);
  void amux_matrix (int, double *, double *, double *, int *, int *);
  void atmux_matrix (int, double *, double *, double *, int *, int *);
  
  void ilut_prec (int, double *, int *, int *, int, double, double *, 
		  int *, int *, int, double *, int *, int *);
  void ilutp_prec(int, double *, int *, int *, int, double, double, int,
		  double *, int *, int *, int, double *, int *, int *,
		  int *);

  void ilud_prec (int, double *, int *, int *, double, double, double *, 
		  int *, int *, int, double *, int *, int *);
  void iluk_prec (int, double *, int *, int *, int, double *, int *, 
		  int *, int *, int, double *, int *, int *);
  void ilu0_prec (int, double *, int *, int *, double *, int *, int *, 
		  int *, int *);
  void msrcsr_format(int, double *, int *, double *, int *, int *, 
		     double *, int *);
  void csrmsr_format (int, double *, int *, int *, double *, int *, 
		      double *, int *);
  void lusol_prec (int, double *, double *, double *, int *, int *);
  void lutsol_prec (int, double *, double *, double *, int *, int *);
  void rungmres_solver(int, double *, double *, int *, double *, 
		       double *, double *, double *, int *, int *, 
		       double *, int *, int *);
  void runfom_solver(int, double *, double *, int *, double *, double *, 
		     double *, double *, int *, int *, double *, int *, 
		     int *);
  void runcg_solver(int, double *, double *, int *, double *, double *, 
		    double *, double *, int *, int *, double *, int *, 
		    int *);
  void runbcg_solver(int, double *, double *, int *, double *, double *, 
		     double *, double *, int *, int *, double *, int *, 
		     int *);
  void rundbcg_solver(int, double *, double *, int *, double *, double *, 
		      double *, double *, int *, int *, double *, int *, 
		      int *);
  void runbcgstab_solver(int, double *, double *, int *, double *, 
			 double *, double *, double *, int *, int *, 
			 double *, int *, int *);
  void runtfqmr_solver(int, double *, double *, int *, double *, double *, 
		       double *, double *, int *, int *, double *, int *, 
		       int *);
  void runfgmres_solver(int, double *, double *, int *, double *, double *, 
			double *, double *, int *, int *, double *, int *, 
			int *);
  void runcgnr_solver(int, double *, double *, int *, double *, double *, 
		      double *, double *, int *, int *, double *, int *, 
		      int *);
  void rundqgmres_solver(int, double *, double *, int *, double *, 
			 double *, double *, double *, int *, int *, 
			 double *, int *, int *);
  
  void multic_matrix (int, int *, int *, int *, int *, int *, int *, int , 
		      int *);

  void indset0_matrix (int, int *, int *, int *, int *, int *, int, int);
  void indset1_matrix (int, int *, int *, int *, int *, int *, int *, int, 
		       int);
  
  void bandwidth(int, int *, int *, int *, int *, int *, double *);
  void nonz(int, int, int *, int *, int *, int *, int *, int *, int *, int *, int *);
  void frobnorm(int, int, double *, int *, int *, double *);
  void ansym(int, int, double *, int *, int *, double *, int *, int *, int *, double *, double *, double *);
  void distaij(int, int, int, int *, int *, double *, double *);
  void skyline(int, int, int *, int *, int *, int *, int *);
  void distdiag(int, int, int *, int *, int *);
  void bandpart(int, int *, int *, int *, int, int *);
  
  void diagdomi_(int*, int*, int*, double *, int *, int *, double *, int *, int *, double *, double *);
  void nimpdiag_(int*, int*, int *, int*, int *, int *, double *);
  void nonzlud_(int*, int *, int *, int *, int *, int *);
  void avnzcol_(int*, int *, int *, int *, int*, double *, double *);
  
#define diag_domi(a, b, c, d, e, f, g, h, i, j, k) diagdomi_(a, b, c, d, e, f, g, h, i, j, k)
#define n_imp_diag(a, b, c, d, e, f, g) nimpdiag_(a, b, c, d, e, f, g)
#define nonz_lud(a, b, c, d, e, f) nonzlud_(a, b, c, d, e, f)
#define avnz_col(a, b, c, d, e, f, g) avnzcol_(a, b, c, d, e, f, g)

  void csrcsc(int, int, int, double *, int *, int *, double *, int *, int *);
}

#endif

