%{

//////////////////////////////////////////////////////////
//
// spmath.l
// ~~~~~~~~
//
//  This is the lex definition file for spmath.  This file
// is intended to be input for gnu's flex which will turn
// it into a C file that contains the code for spmath's
// lexacagraphical analyzer.  The analyzer (lexer) will
// take text as input and output a "token stream" that
// represents the input but tokenized into a more
// comprehensibe form for the parser.  For example...
//
// 3 + 3
//
// ...is turned into...
//
// NUMBER PLUS NUMBER
//
// Where 'NUMBER' and 'PLUS' are numerical constants. In
// the case that the token is not specific (for example,
// 'PLUS' specifically means the operator +, where
// 'NUMBER' does not refer to a specific number) the extra
// needed information is stored in a special lex variable,
// yylval of type YYSTYPE, defined in spmath.y to be:
//
// %union {
//   char *name;
//   value *val;
//   list *set;
// }
//
//  This file is divided into three sections which are
// separated by the "%%" symbol.  This section is the
// first section which contains various lex directives and
// a block of C code contained within the "%{" and "%}"
// symbols.
//
//  The second section contains the lex rules, which
// consist of regular expressions and C-coded rules
// which tell the lexer what to do when that regular
// expression is encountered.
//
//  The third section contains extra C-code, which is 
// copied verbatim into the output file.  For more 
// information on lex and yacc definition files, see a 
// good lex&yacc reference.
//
//  To add new tokens, put a %token directive in
// spmath.y's first section and then run bison on it in
// order to generate a new y.tab.h file.  Then the token
// can be used by the lexer.  For example, if you wanted
// to add a new command 'add_one', you would add:
//
// %token ADD_ONE
//
// ...to the first section in spmath.y, then run...
//
// bison -y -d spmath.y
//
// ...then add a line to the second section of this
// file like this...
//
// add_one   { return ADD_ONE; }
//
//////////////////////////////////////////////////////////

#include <string.h>
#include <math.h>
#include "spmath.h"
#include "value.h"
#include "symtab.h"
#include "list.h"
#include "token.h"

#undef  YY_INPUT
#define YY_INPUT(b, r, ms) (r = my_yyinput(b, ms))

#define YY_DECL int yylex YY_PROTO(( YYSTYPE *yylval ))

#define IN_MODE(x) ((yy_start/2 - 1) == x)

int my_yyinput (char *, int);

int yy_start_temp;

extern int clear_alias; // should allow to clear an alias
extern char * str;
extern int str_flag;
extern int text_flag;
extern int capture_text;
extern int nest_count;
extern symtab * ali;
extern symtab * syms;
extern int several_init;

 /* Define the lex start states */

#define CATCH_MODE 1
#define VECTOR_MODE 2
#define NO_NEST_MODE 3

%}

 /* Define two new lex "modes." */

 /* 'catch' must be first and 'vector' must be second!
   If they are changed, the change must be reflected
   in the CATCH_MODE and VECTOR_MODE constants */

%s catch
%s vector
%s no_nest

%%

 /* Second section */

 /* The catch-all state.  When the lexer is in the catch
    state (from a BEGIN CATCH statement), then it will
    read text until it sees a dollar sign, then return 
    it (unless it is in text mode, in which case it does
    nothing). */

<catch>[^$]*"$"	{ 

  /* If we're in text mode... */
  if (text_flag) {

    int i;
    int len    = strlen (yytext);
    char *temp = strdup (yytext);

    //    cout << "Caught " << yytext << " putting back" << endl;

    /* Put it all back so that it can be left as
       text instead of being parsed */
    for (i = len-1; i >= 0; i -= 1) {
      unput (*(temp+i));
    }

    /* Return nothing */
    yylval->name = strdup (""); 
    BEGIN INITIAL;
    return CATCH; 
  }
  else {

    //    cout << "Caught " << yytext << endl;

    /* Return this chunk of code so that it can
       be saved for execution later */
    yylval->name = strdup (yytext); 
    BEGIN INITIAL;
    return CATCH;
  }
}

 /***********************************************************************/
 /*	Key Words							*/
 /***********************************************************************/
break			{ return BREAK; }
continue		{ return CONTINUE; }
else			{ return ELSE; }
elseif			{ return ELSEIF; }
<no_nest>end            { return END; }
end			{ nest_count -= 1; unput (' '); return END; }
endfunction		{ return ENDFUNCTION; }
<no_nest>for            { return FOR; }
for			{ nest_count += 1; return FOR; }
<no_nest>function       { return FUNCTION; }
function 		{ nest_count += 1; return FUNCTION; }
global			{ return GLOBAL; }
<no_nest>if             { return IF; }
if			{ nest_count += 1; return IF; }
return			{ return RETURN; }
<no_nest>while          { return WHILE; }
while			{ nest_count += 1; return WHILE; }
quit|exit		{ sp_quit(0); }
^!			{ return HISTORY; }
history			{ return HIST_LIST; }

 /***********************************************************************/
 /*	Arithmetic Operators						*/
 /***********************************************************************/

"+"|".+"		{ return PLUS; }
"-"|".-"		{ return MINUS; }
"*"			{ return MULT; }
".*"			{ return EMULT; }
"/"			{ return DIV; }
"./"			{ return EDIV; }
"\\"			{ return LDIV; }
".\\"			{ return ELDIV; }
"^"|"**"		{ return POW; }
".^"|".**"		{ return EPOW; }
"'"|".'"		{ return TRANSP; }
"++"			{ return INC; }
"--"			{ return DEC; }
"="			{ return ASSIGN; }

 /***********************************************************************/
 /*	Comparison Operators						*/
 /***********************************************************************/

"<"			{ return LT; }
"<="			{ return LE; }
"=="			{ return EQ; }
">="			{ return GE; }
">"			{ return GT; }
"!="|"~="|"<>"		{ return NE; }

 /***********************************************************************/
 /*	Logical Operators						*/
 /***********************************************************************/

"&"|"&&"		{ return AND; }
"|"|"||"		{ return OR; }
"~"|"!"			{ return NOT; }
";"			{ return SEMI; }
","			{ return COMMA; }
":"			{ return COLON; }

 /***********************************************************************/
 /*	Special Vector Handling						*/
 /***********************************************************************/

 /* Make sure spaces are interpretted as commas (to be like matlab) */

 "["[ ]* { 

  // If we're in catch state, then never mind
  if(yy_start != IN_MODE(CATCH_MODE)) {
    BEGIN vector; 
  }
  return '[';
 }
 <vector>[ ]*;[ ]* {
   return SEMI;
 }
 <vector>[ ]*,[ ]* {
   return COMMA;
 }
 <vector>[ ]*"]" { 
   BEGIN INITIAL; 
   return ']'; 
 }
 <vector>[ ]+ { return COMMA; }

 /***********************************************************************/
 /*	Function Names							*/
 /***********************************************************************/

type                    { return TYPE; }
who                     { return WHO;  }
whos                    { return WHOS; }

demo                    { return DEMO; }
load			{ return LOAD; }
hbout			{ return PRTMT; }
hbin			{ return READMT; }
fdmat			{ return GEN57PT; }
submat			{ return SUBMAT; }
extract			{ return EXTRACT; }
extract_row		{ return EXTRACT_ROW; }
extract_col		{ return EXTRACT_COL; }
filter			{ return FILTER; }
csort			{ return CSORT; }
rmdia			{ return RMDIA; }
getdia			{ return GETDIA; }
getu			{ return GETU; }
getl			{ return GETL; }
getbwd			{ return GETBWD; }
getbwdu			{ return GETBWDU; }
getbwdl			{ return GETBWDL; }
blkfnd			{ return BLKFND; }
cnrms			{ return CNRMS; }
coscal			{ return COSCAL; }
roscal			{ return ROSCAL; }
addblk			{ return ADDBLK; }
get1up			{ return GET1UP; } 
rperm			{ return RPERM; }
cperm			{ return CPERM; }
dperm			{ return DPERM; }
multic			{ return MULTIC; }
colors			{ return COLORS; }
colors2 		{ return COLORS2; }
colorset		{ several_init = 1; return COLORSET; }
indset0			{ return INDSET0; }
indset1			{ return INDSET1; }

view|show               { return SPY; }

solve|solve_system      { return SOLVE_SYSTEM; }

cos			{ return COS; }
sin			{ return SIN; }
tan			{ return TAN; }
cot			{ return COT; }
sec			{ return SC; }
csc			{ return CSC; }

acos			{ return ARCCOS; }
asin			{ return ARCSIN; }
atan			{ return ARCTAN; }
acot			{ return ARCCOT; }
asec			{ return ARCSEC; }
acsc			{ return ARCCSC; }

abs			{ return ABS; }
exp			{ return EXP; }
log			{ return LOG; }
log10			{ return LOG10; }
sqrt			{ return SQRT; }
rand			{ return RND; }
irand			{ return IRND; }

ceil			{ return CEIL; }
floor			{ return FLOOR; }
round			{ return ROUND; }
trunc			{ return TRUNC; }

nnz			{ return NNZ; }
ncol			{ return NCOL; }
nrow			{ return NROW; }
size			{ return SIZE2; }

eye			{ return EYE; }
zeros|zeroes	        { return ZEROS; }
spzeros|spzeroes        { return SPZEROS; }
ones                    { return ONES; }
constmat		{ return INIT; }


input			{ return INPUT; }
clear			{ clear_alias = 1; return CLEAR; }

dense			{ return DENSIFY; }
sparse			{ return SPARSIFY; }

is_dense		{ return ISDENSE; }
is_sparse		{ return ISSPARSE; }
is_string		{ return ISSTRING; }
is_number		{ return ISNUMBER; }
is_vector		{ return ISVECTOR; }
what_is			{ return WHAT_IS; }

alias			{ return ALIAS; }

help|man|"?" { 

  if(nest_count > 0) {
    sp_error("Cannot use help command in a nested construct\n");
    return ERROR;
  }

  BEGIN NO_NEST_MODE;

  return HELP;
}

 /***********************************************************************/
 /*	Other stuff							*/
 /***********************************************************************/

 /* Detect identifiers */
 
[a-zA-Z_][a-zA-Z0-9_]*	{ 
  
  if (clear_alias == 0) {
    if (ali->contains(yytext)) {

      symtab *s = (new token())->get_tok();
      char *temp = new char[100];
      strcpy(temp, ((ali->lookup(yytext)).get_value()).get_str());
      
      if (s->contains(temp)) {
	entry e = s->lookup(temp);
	return ( (int)(e.get_value()).get_num()  );
      }
      else { 
	yylval->name = strdup(yytext);  
	return NAME; 
      }	
    }
    else { 	
      if (several_init) { 
	if (!syms->contains(strdup(yytext))) 
	  syms->insert(entry(strdup(yytext),*(new value(0))));
      }
      yylval->name = strdup (yytext); 
      return NAME;
    }
  }
  else {	
    yylval->name = strdup(yytext);
    return NAME;
  }
}

 /* Detect integers */

[0-9]+ { 
  if (text_flag) { 
    yylval->val = new value(strdup(yytext));
  }
  else 
    yylval->val = new value ((atoi(yytext)));
  return NUMBER; 
} 

 /* Detect floating point numbers */

([0-9]+|([0-9]*\.[0-9]+))([eE][+-]?[0-9]+)? { 
  if (text_flag)
    yylval->val = new value (strdup(yytext));
  else
    yylval->val = new value ((atof(yytext)));
  return NUMBER;
}

 /* Detect shell commands */

\\[^$\n]* { 
  yylval->name = strdup (yytext); 
  return SYSTEM;
}

 /* Reject malformed numbers */

[0-9]+[a-zA-Z_][0-9]* {
  sp_error("Error, malformed integer: ");
  sp_error(yytext);
  sp_error("\n");

  return ERROR;
}

 /* Reject malformed strings */

\"[^"\n]*[\n] {
  sp_error("Error, malformed string: ");
  sp_error(yytext);
  sp_error("\n");

  return ERROR;
}

 /* Detect quoted strings */

\"[^"\n]*["] { 
  yylval->val = new value (strdup (yytext));
  return STR; 
}

#[^$\n]*		; /* Ignore comments. */
[ \t]+			; /* Ignore white space */

 /* Detect end-of-input */

"$"			{ BEGIN INITIAL; return 0; }

 /* Report any character that doesn't fit into any of these rules */

.|\n			{ return yytext[0]; }

%%

    /* Third section */

 /* This is a helper function that allows the static variable
 state to be changed from the parser */

void my_yybegin(int state) {
   BEGIN state;
}

int my_yyinput (char *buf, int max_size) {

  int result;
  int n;

  // If we're lexing a string...
  if (str_flag) {
    n = strlen(str);

    if (n > max_size)
      n = max_size;

    strncpy (buf, str, n+1);
    str_flag = 0;

    result = n;
  } // end if
  else {

    if (yy_current_buffer->yy_is_interactive) {
      
      int c = getc( yyin ); 
      result = c == EOF ? 0 : 1; 
      buf[0] = (char) c; 
    } 

    else if ( ((result = fread( buf, 1, max_size, yyin )) == 0) 
              && ferror( yyin ) ) 
      YY_FATAL_ERROR( "input in flex scanner failed" );

  } // end else  

  return result;
} // end my_yyinput






























