#
# spmath.tcl
# ~~~~~~~~~~
#
#   This script is designed to act as a graphical front-end for the
# sparse matrix toolkit SPMATH.
#
# Author: Ben Langmead (btl10@columbia.edu)
# Date:   July 3, 2001
#

# Check the existance of the support source files and include them

# puts "Hello"

set spmathpath $argv0
set mypath [string range $spmathpath 0 [expr [string last "spmath" $spmathpath] - 1]]

string trimright $mypath "/"

if {$mypath == ""} then {
    set mypath "."
}

# puts $mypath

if [file exists "$mypath/procs.tcl"] {
    source "$mypath/procs.tcl"
    puts {Found file [procs.tcl]...}
} else {
    puts "Cannot find source file procs.tcl, quitting..."
    exit
}

if [file exists "$mypath/view.tcl"] {
    source "$mypath/view.tcl"
    puts {Found file [view.tcl]...}
} else {
    puts "Cannot find source file view.tcl, quitting..."
    exit
}

puts "Finished loading supporting files"

# Set attributes for the main window

#wm geometry . 600x435+5+25
wm title    . {SPMATH Command Interpreter - v0.2}


#
# Create a text widget and pack it
#
set scroll_width 15

frame .sidebar

frame .sidebar.spacer \
	-width [expr $scroll_width + 5] \
	-height [expr $scroll_width + 5]

scrollbar .sidebar.y \
	-orient vertical \
	-command { .t yview } \
	-width $scroll_width

scrollbar .x \
	-orient horizontal \
	-command { .t xview } \
	-width $scroll_width

text .t \
	-wrap none \
	-background white \
	-highlightcolor white \
	-highlightthickness 0 \
	-tabs { 2c left } \
	-spacing1 0 \
	-spacing3 0 \
	-height 33 \
	-width 80 \
	-takefocus 1 \
	-yscrollcommand { .sidebar.y set } \
	-xscrollcommand { .x set }
bindtags .t {MyText .t Text all}

.t insert end " "

#set ch_wd [lindex [.t bbox "insert -1 chars"] 2]
#set ch_ht [lindex [.t bbox "insert -1 chars"] 3]

pack .sidebar        -side right  -fill y
pack .sidebar.spacer -side bottom
pack .sidebar.y      -side right  -fill y
pack .x              -side bottom -fill x
pack .t              -side right  -fill both -expand 1

#
# Configure the tags so that they alter the appearance of the text that is
# tagged with them.  Notice that the earlier tags in the taglist take 
# precedence.  I.e. the multiple tagged example will have the display
# attributes of the first tag, but the text will respond to any of the 
# bindings on any of its tags.
#

.t tag configure result_tag \
	-foreground blue \
	-lmargin1 5 \
	-lmargin2 5 \
	-spacing1 0 \
	-spacing3 0

.t tag configure error_tag \
	-foreground red \
	-lmargin1 5 \
	-lmargin2 5 \
	-spacing1 0 \
	-spacing3 0

# The prompt must be 3 characters long in order for the "return"
# bounded script to be able to retrive the command properly

set prompt ">  "
set nest_prompt "-> "

set autoscroll 1
set command_num 1
set history_num 1
set mouse_down 0
set nested 0

# Give main text box the focus

focus .t

#
# Window lifecycle events
#

bind . <Destroy> {
    continue
}

#
# Keyboard events
#

bind MyText <Key> {
    
    set ins_pos [.t index insert]
    set end_pos [.t index end]
    if [.t compare $ins_pos < "end -1 lines linestart +3 chars"] then {
	.t mark set insert end
    }
}

bind .t <Key-Up> {

    history_prev

    .t see insert

    break
}

bind .t <Key-Down> {

    history_next

    .t see insert

    break
}

bind .t <Key-Home> {
    .t mark set insert "end -1 lines linestart +3 chars"

    break
}

bind .t <Key-Left> {

    # Check if cursor is running into prompt

    .t see insert

    set ins_pos [.t index insert]
    if [.t compare $ins_pos < "end -1 lines linestart +4 chars"] then {
	break
    }
}

bind .t <Control-a> {
    .t mark set insert "insert linestart +3 chars"
    break
}

bind .t <Control-u> {
    .t delete "insert linestart +3 chars" "insert lineend"
    break
}

bind .t <Control-l> {
    .t delete "0.1" "end"
    .t insert end $prompt

    break
}

bind .t <Control-w> {
    .t delete "insert linestart +3 chars" "insert"
    break
}

bind .t <Control-k> {
    .t delete "insert" "insert lineend"
    break
}

bind .t <Control-f> {
    .t mark set insert "insert +1 chars"
    break
}

bind .t <Control-b> {
    .t mark set insert "insert -1 chars"
    break
}

bind .t <Control-h> {

    # Check to see if cursor is running into prompt
    set ins_pos [.t index insert]
    if [.t compare $ins_pos < "end -1 lines linestart +4 chars"] then {
	.t mark set insert "end -1 lines lineend"
	break;
    }

    .t delete "insert -1 chars" "insert"
    break
}

bind .t <BackSpace> {

    # Check to see if cursor is running into prompt
    set ins_pos [.t index insert]
    if [.t compare $ins_pos < "end -1 lines linestart +4 chars"] then {
	.t mark set insert "end -1 lines lineend"
	break;
    }

    .t delete "insert -1 chars" "insert"
    break
}

bind MyText <Delete> {

    # Check to see if cursor is running into prompt
    set ins_pos [.t index insert]
    set end_pos [.t index end]
    if [.t compare $ins_pos < "end -1 lines linestart +4 chars"] then {
	.t mark set insert "end"
	break;
    }

    # We have to override delete's functionality, otherwise
    # it will delete the selection.

    .t delete "insert" "insert +1 chars"
    break
}

bind .t <Return> {
    .t mark set insert end
 
    # Get the line of text the the user just entered
    set command_str [.t get "insert linestart +3 chars" "insert lineend"]

    # Trim away whitespace
    set command_str [string trim $command_str]
 
    # If it is an empty string, then ignore it
    if [expr [string length $command_str] == 0] {
	.t insert end "\n"
	.t insert end $prompt

	if $autoscroll then {
	    .t see insert
	}

	break
    }

   .t insert end "\n"
   
    # Call the C interface for the parser using command_str
    # as the input for lex/yacc
    get_command $command_str

    # If we're nested, then do not add to history yet and
    # change prompt so that user knows to continue input

#    if $nested then {

#	set prompt "-> "
#	.t insert end $prompt
#    } else {

	# If not nested, make sure prompt is normal and
	# increment the history variables
#	set prompt ">  "

#	.t insert end "\n"
#	.t insert end $prompt

#	set command_num [expr $command_num + 1]
#	set history_num $command_num
#    }

    # Don't let any more handlers take control
    break
}

bind .t <ButtonRelease-2> {
    set cur_pos [.t index current]
    set end_pos [.t index end]

#    if {[.t tag ranges sel] != ""} then {
	if [.t compare $cur_pos < "end -1 lines linestart +3 chars"] then {
#	    .t mark set insert end
#	    continue
#	    .t insert end [.t get "sel.first" "sel.last"]
	    .t mark set insert end
	    catch {%W insert insert [selection get -displayof %W]}
	    if [string match [%W cget -state] normal] {focus %W}
	    break
	} else {
	    continue
	}
#    } else {
#	puts "No selection"
#    }
}






