//////////////////////////////////////////////////////////
//
// tcl.C
// ~~~~~
//  This file defines all the C++ functions that are 
// called from tcl/tk in order to interface with spmath.
//
// Author: Ben Langmead (btl10@columbia.edu)
// Date:   June 29, 2001
//
//////////////////////////////////////////////////////////

#ifndef NO_TK_COMPILE

#include <string.h>
#include <stdlib.h>
#include "spmath.h"

// Useful typedef for tk
extern "C" { typedef void (*vfp)(void *); }

// the C commands to be callable from tcl

extern "C" {

int get_command_proc  (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int history_prev_proc (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int history_next_proc (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int mouse_motion_proc (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int zoom_out_proc     (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int view_close_proc   (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int view_resize_proc  (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int view_zoom_proc    (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int view_recolor_proc (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int view_stats_proc   (ClientData cd, Tcl_Interp *itp, int argc, char **argv);
int view_redraw_proc  (ClientData cd, Tcl_Interp *itp, int argc, char **argv);

}

//
// create_tcl_commands
//
//  Called from tcl_init in main.C in order to register 
// these C++ functions with the tcl/tk interpretter.
//

void create_tcl_commands() {

  Tcl_CreateCommand( interp, strdup("get_command"),
		     get_command_proc,  (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("history_prev"),
		     history_prev_proc, (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("history_next"),
		     history_next_proc, (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("mouse_motion"),
		     mouse_motion_proc, (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("zoom_out"),
		     zoom_out_proc,     (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("view_close"),
		     view_close_proc,   (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("view_resize"),
		     view_resize_proc,  (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("view_zoom"),
		     view_zoom_proc,    (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("view_recolor"),
		     view_recolor_proc, (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("view_stats"),
		     view_stats_proc,   (ClientData) NULL, (vfp)NULL );

  Tcl_CreateCommand( interp, strdup("view_redraw"),
		     view_redraw_proc,  (ClientData) NULL, (vfp)NULL );
}

//
// get_command_proc
//
//  This function is called from the main text widget
// when the user presses enter at the end of a command.
// The text entered is passed to this function in
// argv[1].  This function will also set a gobal variable
// "nested" in the tcl/tk code if the command is an
// unfinished nested construct.
//

int get_command_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {

  if( argc != 2 ) {

    Tcl_SetResult( itp, strdup("usage: get_command_proc <command string>"),
		   TCL_STATIC );

    return( TCL_ERROR );
  }
  if( strcmp(argv[1], "") ) {

    if(sp_parse((char *)strdup(argv[1]))) {
      TCL("set nested 1");
    }
    else {
      TCL("set nested 0");
    }
  }

  return( TCL_OK );
}

//
// history_next_proc
//
//  This function is called from the main text widget
// when the user presses the up or down buttons to
// request a previous command from the history list.
//

int history_next_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {

  if( argc != 1 ) {

    Tcl_SetResult( itp, strdup("usage: history_next"),
		   TCL_STATIC );

    return(TCL_ERROR);
  }

  hist_cur++;
  if(hist_cur >= hist_num) {
    hist_cur = hist_num;
    sp_command("");

    return(TCL_OK);
  }

  char * str = hist_list.get_item(hist_cur).get_pointer();
  *(str + strlen(str) - 1) = '\0';
  
  if(strcmp(str, "")) {
    sp_command(str);
  }

  return(TCL_OK);
}

//
// Outputs the previous history item
//
int history_prev_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {

  if( argc != 1 ) {

    Tcl_SetResult( itp, strdup("usage: history_prev"),
		   TCL_STATIC );

    return(TCL_ERROR);
  }

  hist_cur--;
  if(hist_cur < 1) {
    hist_cur = 1;
  }

  char * str = hist_list.get_item(hist_cur).get_pointer();
  *(str + strlen(str) - 1) = '\0';
  
  if(strcmp(str, "")) {
    sp_command(str);
  }

  return(TCL_OK);
}

//
// mouse_motion_proc
//
//  Called from a sparse matrix viewing window when the
// user's cursor moves over the image of the matrix.
//

int mouse_motion_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {

  int wndnum = 0, x = 0, y = 0, ms = 0;

  if(argc != 5) {
    Tcl_SetResult(itp, strdup("usage: mouse_motion <window> <x> <y> <mousedown>"), TCL_STATIC);
    return TCL_ERROR;
  }

  if(strcmp(argv[1], "")) {
    // Chop the number off the end of the string
    char * wnd = argv[1] + (strlen(argv[1]) - 3);
    wndnum = atoi(wnd);
  } else { return TCL_ERROR; }

  if(strcmp(argv[2], "")) {
    x = atoi(argv[2]);
  } else { return TCL_ERROR; }

  if(strcmp(argv[3], "")) {
    y = atoi(argv[3]);
  } else { return TCL_ERROR; }

  if(strcmp(argv[4], "")) {
    ms = atoi(argv[4]);
  } else { return TCL_ERROR; }

  if(spyviews[wndnum] != NULL) {
    spyviews[wndnum]->mouse_track(x, y, ms);
  }

  return TCL_OK;
}

//
// zoom_out_proc
//
//  Called from a sparse matrix viewing window when
// the user requests for the view to be zoomed out.
//

int zoom_out_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {
  int wndnum = 0;

  if(argc != 2) {
    Tcl_SetResult(itp, strdup("usage: zoom_out <window>"), TCL_STATIC);
    return TCL_ERROR;
  }

  if(strcmp(argv[1], "")) {
    // Chop the number off the end of the string
    char * wnd = argv[1] + (strlen(argv[1]) - 3);
    wndnum = atoi(wnd);
    if(spyviews[wndnum] != NULL) {

      spyviews[wndnum]->zoom_out();
      return TCL_OK;
    }
  }

  return TCL_ERROR;
}

int view_close_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {
  int wndnum = 0;

  if(argc != 2) {
    Tcl_SetResult(itp, strdup("usage: view_close <window>"), TCL_STATIC);
    return TCL_ERROR;
  }

  if(strcmp(argv[1], "")) {
    // Chop the number off the end of the string
    char * wnd = argv[1] + (strlen(argv[1]) - 3);
    wndnum = atoi(wnd);
    if(spyviews[wndnum] != NULL) {

      // Dispose of the view object
      delete spyviews[wndnum];
      spyviews[wndnum] = NULL;

      return TCL_OK;
    }
  }

  return TCL_ERROR;
}

int view_resize_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {
  int wndnum = 0, wd, ht, ix, iy;

  if(argc != 6) {
    Tcl_SetResult(itp, strdup("usage: view_resize <window> <width> <height> <xoff> <yoff>"), TCL_STATIC);
    return TCL_ERROR;
  }

  if(strcmp(argv[1], "")) {
    // Chop the number off the end of the string
    char * wnd = argv[1] + (strlen(argv[1]) - 3);
    wndnum = atoi(wnd);
  }

  if(strcmp(argv[2], "")) {
    wd = atoi(argv[2]);
  } else { return TCL_ERROR; }

  if(strcmp(argv[3], "")) {
    ht = atoi(argv[3]);
  } else { return TCL_ERROR; }

  if(strcmp(argv[4], "")) {
    ix = atoi(argv[4]);
  } else { return TCL_ERROR; }

  if(strcmp(argv[5], "")) {
    iy = atoi(argv[5]);
  } else { return TCL_ERROR; }

  if(spyviews[wndnum] != NULL) {
    spyviews[wndnum]->resize(wd, ht, ix, iy);
  }

  // Don't return an error if spyviews[wndnum] is NULL,
  // because this sometimes will be called from the 
  // view's constructor before it becomes non-NULL

  return TCL_OK;
}

int view_zoom_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {
  int wndnum = 0, zoom = 0;

  if(argc != 3) {
    Tcl_SetResult(itp, strdup("usage: view_zoom <window> <zoom percentage>"), TCL_STATIC);
    return TCL_ERROR;
  }

  if(strcmp(argv[1], "")) {
    // Chop the number off the end of the string
    char * wnd = argv[1] + (strlen(argv[1]) - 3);
    wndnum = atoi(wnd);
  } else { return TCL_ERROR; }

  if(strcmp(argv[2], "")) {
    zoom = atoi(argv[2]);
  } else { return TCL_ERROR; }

  if(spyviews[wndnum] != NULL) {
    spyviews[wndnum]->zoom(zoom);
  }

  return TCL_OK;
}

int view_recolor_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {
  int wndnum = 0;
  unsigned int nr, ng, nb, pr, pg, pb;

  if(argc != 4) {
    Tcl_SetResult(itp, strdup("usage: view_zoom <window> <neg> <pos>"), TCL_STATIC);
    return TCL_ERROR;
  }

  if(strcmp(argv[1], "")) {
    // Chop the number off the end of the string
    char * wnd = argv[1] + (strlen(argv[1]) - 3);
    wndnum = atoi(wnd);
  } else { return TCL_ERROR; }

  if(strcmp(argv[2], "")) {
    sscanf(argv[2], "#%2x%2x%2x", &nr, &ng, &nb);
  } else { return TCL_ERROR; }

  if(strcmp(argv[3], "")) {
    sscanf(argv[3], "#%2x%2x%2x", &pr, &pg, &pb);
  } else { return TCL_ERROR; }

  if(spyviews[wndnum] != NULL) {
    spyviews[wndnum]->recolor((unsigned char)nr,
			      (unsigned char)ng,
			      (unsigned char)nb,
			      (unsigned char)pr,
			      (unsigned char)pg,
			      (unsigned char)pb);
  }

  return TCL_OK;
}

int view_stats_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {
  int wndnum = 0;

  if(argc != 2) {
    Tcl_SetResult(itp, strdup("usage: view_stats <window>"), TCL_STATIC);
    return TCL_ERROR;
  }

  if(strcmp(argv[1], "")) {
    // Chop the number off the end of the string
    char * wnd = argv[1] + (strlen(argv[1]) - 3);
    wndnum = atoi(wnd);
  } else { return TCL_ERROR; }

  if(spyviews[wndnum] != NULL) {
    spyviews[wndnum]->set_stats();
  }

  return TCL_OK;
}

int view_redraw_proc(ClientData cd, Tcl_Interp *itp, int argc, char **argv) {
  int wndnum = 0;

  if(argc != 2) {
    Tcl_SetResult(itp, strdup("usage: view_redraw <window>"), TCL_STATIC);
    return TCL_ERROR;
  }

  if(strcmp(argv[1], "")) {
    // Chop the number off the end of the string
    char * wnd = argv[1] + (strlen(argv[1]) - 3);
    wndnum = atoi(wnd);
  } else { return TCL_ERROR; }

  if(spyviews[wndnum] != NULL) {
    spyviews[wndnum]->draw();
  }

  return TCL_OK;
}

#endif
