#!/usr/local/bin/wish4.1

#################################################################
#
# view.tcl
# ~~~~~~~~
#
#   This file is included from spmath.tcl and encapsulates the
# functionality of SPMATH's Sparse Matrix Viewer.  Its procedures
# are invoked from view.C in order to handle the tcl/tk layout
# of the window while view.C handles the drawing of the image.
#
# Author: Ben Langmead (btl10@columbia.edu)
# Date:   July 6, 2001
#
#################################################################

#
# Globals
# ~~~~~~~

set mouse_is_down 0
set mouse_is_in_bounds 0

set zoom_is_showing 0

set tfont     "-*-courier-medium-r-normal--*-*-*-*-*-*"

set image_wd 300
set image_ht 300

# These strings are read by the C++ code

set image_path   {s.img}
set color_spread_path {color.img}

set zoom_percentages [list {95} {90} {85} {75} {50} {25} {00}]
set zoom_selection "none"

#
# Canvas event handlers
# ~~~~~~~~~~~~~~~~~~~~~

# Called when the mouse moves over a canvas
bind ImageCanvas <Motion> {
    set par [winfo parent %W]

    # If the mouse is somewhere over the canvas, then
    # call the mouse_track method to draw its rectangle
    
    if $mouse_is_in_bounds then {
	mouse_motion $par %x %y $mouse_is_down
    }
}

# Called when a canvas is resized...
# Usually, the canvas is resized only when the parent
# window is resized.  This happens because the canvas
# is "packed" using the option "-expand 1".
bind ImageCanvas <Configure> {
    set par [winfo parent %W]

    resize $par %w %h %x %y
}

# Called when the mouse cursor enters a canvas' region
bind ImageCanvas <Enter> {
    global mouse_is_in_bounds
    set mouse_is_in_bounds 1
}

# Called when the mouse leaves a canvas' region
bind ImageCanvas <Leave> {
    global mouse_down
    global mouse_is_in_bounds
    global zoom_is_showing

    set mouse_is_in_bounds 0
    set par [winfo parent %W]

    %W delete rect
    $par.bottom.xcoord configure -text {X:}
    $par.bottom.ycoord configure -text {Y:}
    $par.bottom.val configure -text {Value:}

    break
}

# Called when someone clicks in the canvas region
bind ImageCanvas <Button-1> {
    set par [winfo parent %W]
    global mouse_down

    set mouse_is_down 1

    mouse_motion $par %x %y $mouse_is_down

    break
}

# Called when someone releases the mouse button within
# the canvas region
bind ImageCanvas <ButtonRelease-1> {
    set par [winfo parent %W]

    global mouse_down

    set mouse_is_down 0

    mouse_motion $par %x %y $mouse_is_down

    break
}

# Called when the right mouse button is clicked within
# a canvas region
bind ImageCanvas <Button-3> {
    set par [winfo parent %W]

    zoom_out $par
}

#
# ZoomButton event handlers
# ~~~~~~~~~~~~~~~~~~~~~~~~~

# Called when the Zoom... button is pressed
bind ZoomButton <Button-1> {

    global zoom_is_showing

    set par [winfo parent [winfo parent %W]]

    # If percentages are showing, hide them and vice versa
    if {$zoom_is_showing} then {
	hide_zoom_menu $par
    } else {
	show_zoom_menu $par
    }
}

# These ensure that the button still lights up when the
# cursor moves over it (otherwise this behavior would be
# overridden by ZoomButton's bindtags)
bind ZoomButton <Enter> { %W configure -state active }
bind ZoomButton <Leave> { %W configure -state normal }

#
# ZoomLabel event handlers
# ~~~~~~~~~~~~~~~~~~~~~~~~

# Called when the mouse moves over a zoom percentage 
# label
bind ZoomLabel <Enter> {
    global zoom_selection

    set par [winfo parent %W]

    # Flatted the previous label and raise this one
    if {$zoom_selection != "none"} then {
	$zoom_selection configure -relief flat
    }

    set zoom_selection %W
    %W configure -relief raised
}

# Called when the mouse leaves a zoom percentage label
bind ZoomLabel <Leave> {
    # Make it flat
    %W configure -relief flat
    set zoom_selection "none"
}

# Called when the mouse is clicked over a percentage
# label
bind ZoomLabel <Button1-ButtonRelease-1> {
    global zoom_selection

    set par [winfo parent [winfo parent %W]]

    if {$zoom_selection != "none"} then {
	
	# Flatten the label and hide the panel
	$zoom_selection configure -relief flat
	set zoom_selection "none"
	hide_zoom_menu $par

	# Get the number percentage and call the C++
	# zoom method via zoom
	set zl [string trimright [lindex [%W configure -text] 4] {%%}]
	zoom $par $zl
    }
}


bind NegativeColor <Button-1> {
    %W configure -relief sunken
}

bind NegativeColor <Button1-ButtonRelease-1> {
    set par [winfo parent [winfo parent %W]]
    
    %W configure -relief raised

    set cur [lindex [%W configure -background] 4]
    set new [tk_chooseColor -initialcolor $cur -parent . -title "Choose Negative Color"]

    if {$new != ""} then {
	%W configure -background $new
	recolor $par
	redraw $par
    }    
}


bind PositiveColor <Button-1> {
    %W configure -relief sunken
}

bind PositiveColor <Button1-ButtonRelease-1> {
    set par [winfo parent [winfo parent %W]]
    
    %W configure -relief raised

    set cur [lindex [%W configure -background] 4]
    set new [tk_chooseColor -initialcolor $cur -parent . -title "Choose Positive Color"]

    if {$new != ""} then {
	%W configure -background $new
	recolor $par
	redraw $par
    }
}

#
# Toplevel event handlers
# ~~~~~~~~~~~~~~~~~~~~~~~

# Called when a toplevel is in the process of being closed.
bind Toplevel <Destroy> {
    # Call the C++ method that frees the memory taken up
    # by the view (see ~view(), view.C)
    
    view_close %W
}

#
# Procedures
# ~~~~~~~~~~

#
# create_view { view }
#
# Creates a new Sparse Matrix Viewer by creating a new toplevel
# named $view.  Called by view's constructors in view.C.
#

proc create_view { view } {

    # Create new toplevel window named ".spyviewXXX" where
    # "XXX" is some zero-padded number supplied by the 
    # C++ code (see create_tk_window(), view.C)

    toplevel    $view
    wm title    $view {SPMATH Sparse Matrix Viewer}

    toplevel     $view.info
    wm title     $view.info {Sparse Matrix Stats}
    wm resizable $view.info 0 0
    wm withdraw  $view.info
    wm protocol  $view.info WM_DELETE_WINDOW "less_info $view"
    bindtags     $view.info {$view.info InfoToplevel all}

    global tfont
    global image_wd
    global image_ht

    global zoom_percentages

    # Create the image that will be bound to the canvas

    image create photo $view.s.img

    canvas $view.s \
	    -width $image_wd \
	    -height $image_ht \
	    -relief sunken \
	    -borderwidth 0
    bindtags $view.s {$view.s ImageCanvas all}
    
    # Bind the image to the canvas

    $view.s create image 0 0 -image $view.s.img -anchor nw

    # Create all the objects that make up the window...

    frame $view.more
    frame $view.pos
    frame $view.top
    frame $view.bottom
    frame $view.colall -relief sunken -borderwidth 1
    frame $view.color
    frame $view.colorlabels
    frame $view.zoom

    frame $view.zoompop \
	    -relief raised \
	    -width 100 \
	    -height 100 \
	    -borderwidth 3

    frame $view.top.upper
    frame $view.top.lower

    label $view.top.dim  -height 1
    label $view.top.nnz  -height 1
	
    label $view.bottom.xcoord -text {X: } -anchor w -pady 5 -width 10
    label $view.bottom.ycoord -text {Y: } -anchor w -pady 5 -width 10
    label $view.bottom.val -text {Value: } -anchor w -justify left -pady 5
    
    label $view.pos.zoomper -text {Zoom: 0.0%} -anchor w -width 12
    label $view.pos.region -text {Vis. Region: } -anchor e -width 30
    
    frame $view.color.neg \
	    -width 30 \
	    -height 30 \
	    -relief raised \
	    -borderwidth 2 \
	    -background "#FF00FF"
    bindtags $view.color.neg "$view.color.neg NegativeColor all"

    frame $view.color.pos \
	    -width 30 \
	    -height 30 \
	    -relief raised \
	    -borderwidth 2 \
	    -background "#0000FF"
    bindtags $view.color.pos "$view.color.pos PositiveColor all"

    canvas $view.color.spread \
	    -height 28 \
	    -relief sunken \
	    -borderwidth 1 \
	    -background "#FFFFFF"
    bindtags $view.color.spread "$view.color.pos all"

    label $view.colorlabels.neg -text { } -anchor w
    label $view.colorlabels.pos -text { } -anchor e
    label $view.colorlabels.zero -text {Zero} -anchor n -justify center

    image create photo $view.color.img

    $view.color.spread create image 0 0 -image $view.color.img -anchor nw

    button   $view.zoom.zoom -text "Zoom..."
    bindtags $view.zoom.zoom "ZoomButton"

    button $view.zoom.more -command "more_info $view" -text "More info > "
    
    foreach z $zoom_percentages {
	label $view.zoompop.p$z -text "$z%" -anchor w
	bindtags $view.zoompop.p$z "$view.zoompop.p$z ZoomLabel all"
	pack  $view.zoompop.p$z -side top -fill x -padx 3
    }

    label $view.status -text {Loading matrix...} -relief sunken
    label $view.title  -text {} -pady 5 -padx 5
    
    # This text box contains the "More info" matrix information
    # It is filled with text by "set_labels" in view.C

    text $view.info.t \
	    -wrap none \
	    -border 1 \
	    -font $tfont \
	    -width 62 \
	    -height 42 \
	    -tabs 50 \
	    -padx 10 \
	    -pady 10 \
	    -highlightthickness 0

    pack $view.info.t \
	    -fill both \
	    -padx 10 \
	    -pady 10

    # Pack everything into the window
    
    pack $view.bottom.xcoord -side left -padx 5
    pack $view.bottom.ycoord -side left -padx 5
    pack $view.bottom.val    -side left -fill x
    
    pack $view.pos.zoomper -side left -padx 5
    pack $view.pos.region -side right -padx 5

    pack $view.zoom.more -side right
    pack $view.zoom.zoom -side left
    
    pack $view.top.dim -side left -padx 4
    pack $view.top.nnz -side right -padx 4
    
    pack $view.title -side top -fill x
    pack $view.top -side top
   
    pack $view.status -side bottom -padx 5 -pady 5 -fill x
    pack $view.zoom   -side bottom -padx 5 -fill x
    pack $view.color  -side bottom -fill x -in $view.colall
    pack $view.colorlabels -side bottom -fill x -in $view.colall
    pack $view.colall -side bottom -padx 5 -pady 5 -fill x

    pack $view.color.neg -side left
    pack $view.color.pos -side right
    pack $view.color.spread -side left -fill x -expand 1 -ipadx 4

    pack $view.colorlabels.neg -side left
    pack $view.colorlabels.pos -side right
    pack $view.colorlabels.zero -side left -fill x -expand 1

    pack $view.bottom -side bottom -padx 5 -fill x
    pack $view.s      -side bottom -padx 5 -pady 2 -fill both -expand 1
    pack $view.pos    -side bottom -padx 5 -pady 2 -fill x

    # Make sure that the window is displayed before the
    # matrix gets drawn, otherwise the program gives no
    # feedback to the user for the long time that it
    # spends drawing the image...

    wm minsize $view [winfo width $view] [winfo height $view]
}

#
# more_info { view }
#
# Adds a panel with statistical information gathered with
# get_stats in a text box.  Targets view "view".
#

proc more_info { view } {
    # Change "More info" button to "Less info" button

    view_stats $view

    $view.zoom.more configure -text "Less info <"
    $view.zoom.more configure -command "less_info $view"

    wm deiconify $view.info
}

#
# less_info { view }
#
# Takes away the "more info" panel from view "view".
#

proc less_info { view } {
    # Change "Less info" button to "More info" button

    $view.zoom.more configure -text "More info >"
    $view.zoom.more configure -command "more_info $view"

    wm withdraw $view.info
}

#
# draw_rect { view rx1 ry1 rx2 ry2 } 
#
# Draws a rectangle with upper-left hand corner rx1, ry1
# and lower right-hand corner rx2, ry2 in view view.
# Called from mouse_track() method in view.C
#

proc draw_rect { view rx1 ry1 rx2 ry2 } {

    $view.s create rectangle $rx1 $ry1 $rx2 $ry2 -outline black -tag rect
}

#
# delete_view { view }
#
# Deletes the toplevel and image for a Matrix Viewer window
# named $view.  Called by ~view() in view.C
#

proc delete_view { view } {

    if [winfo exists $view.s.img] then {
	image delete $view.s.img
    }
    if [winfo exists $view] then {
	destroy $view
    }
}

#
# resize { view w h x y }
#
# Responds to resizings of a matrix view and redraws the
# image using the new size.
#

proc resize { view w h x y } {
    # Set width and height and subtrack the two-thick
    # border from the image

    set wd [expr $w - 3]
    set ht [expr $h - 3]

    # This is to ensure that the "more" panel does not
    # invade the left panel's space when it's expanded

    $view.top configure -width $wd

    # Set wd, ht, ix and iy to store the width, height,
    # x offset and y offset of the image within the
    # canvas.

    if [expr $w > $h] then {
	set ix [expr (($w - $h) / 2) + 1]
	set iy 1
	set wd $ht
    } else {
	set ix 1
	set iy [expr (($h - $w) / 2) + 1]
	set ht $wd
    }

    # Draw the bordering rectangle

    $view.s delete outer
    $view.s create rectangle $ix $iy \
	    [expr ($ix + $wd)] \
	    [expr ($iy + $ht)] \
	    -outline black \
	    -tag outer

    # Call the View method in order to re-draw the matrix

    view_resize $view $wd $ht $ix $iy

    recolor $view
}

proc redraw { view } {
    view_redraw $view
}

proc recolor { view } {
    $view.color.img configure -width  [winfo width  $view.color.spread]
    $view.color.img configure -height [winfo height $view.color.spread]
    
    view_recolor $view \
	    [lindex [$view.color.neg configure -background] 4] \
	    [lindex [$view.color.pos configure -background] 4]
}

#
# zoom { view perc }
#
# Calls the appropriate C++ method in order to zoom view's
# image to perc percent, while remaining centered.
#

proc zoom { view perc } {
    view_zoom $view $perc
}

#
# show_zoom_menu { view }
#
# Pops up the frame containing the zoom percentages in view
# $view.
#

proc show_zoom_menu { view } {
    global zoomx
    global zoomy
    global zoomwd
    global zoomht
    global zoom_is_showing

    set zoom_is_showing 1

    set w_ht [winfo height $view]

    set b_wd [winfo width $view.zoom.zoom]
    set b_ht [winfo height $view.zoom.zoom]
    set b_x [winfo x $view.zoom]
    set b_y [winfo y $view.zoom]

    set x [expr $b_x + $b_wd]

    set z_ht [winfo height $view.zoompop]

    if {$z_ht == 1} then {
	place $view.zoompop -x $x -y $b_y
	update
	set z_ht [winfo height $view.zoompop]
	place forget $view.zoompop
    }

    set y [expr $b_y]

    if {[expr $y + $z_ht] > $w_ht} then {
	set y [expr $w_ht - $z_ht]
    }

    place $view.zoompop -x $x -y $y
    raise $view.zoompop

    $view.zoom.zoom configure -relief sunken
}

#
# hide_zoom_menu { view }
#
# Make the frame containing the zoom percentages disappear from
# view $view
#

proc hide_zoom_menu { view } {
    global zoom_is_showing

    set zoom_is_showing 0

    $view.zoom.zoom configure -relief raised

    place forget $view.zoompop
}

proc init_view { view } {
    global image_wd
    global image_ht

    recolor $view
    update
}

proc blank_image { view } {
    $view.s.img blank
}

proc clear_rectangles { view } {
    $view.s delete rect
}

proc set_status { view s } {
    $view.status configure -text $s
}

proc set_x { view x } {
    $view.bottom.xcoord configure -text "X: $x"
}

proc set_y { view y } {
    $view.bottom.ycoord configure -text "Y: $y"
}

proc set_value { view v } {
    $view.bottom.val configure -text "Value: $v"
}

proc set_info_text { view text } {
    $view.info.t insert end $text
    $view.info.t configure -state disabled
}

proc set_vis_region { view wd ht } {
    $view.pos.region configure -text [format {Vis. Region: %dx%d} $wd $ht]
}

proc set_zoom_percentage { view p } {
    $view.pos.zoomper configure -text "Zoom: $p"
}

proc set_title { view t } {
    $view.title configure -text $t
}

proc set_dimension { view wd ht } {
    $view.top.dim configure -text [format {Dimensions : %dx%d} $wd $ht]
}

proc set_nnz { view n } {
    $view.top.nnz configure -text "Nonzeros : $n"
}

proc set_spread_limits { view nl pl } {
    $view.colorlabels.neg configure -text $nl
    $view.colorlabels.pos configure -text $pl
}






