//////////////////////////////////////////////////////////
//
// view.C
// ~~~~~~
//
//   The implementation of the view class.  The view class
// is meant to encapsulate the functionality of a sparse
// matrix viewing window.  It interfaces with tcl/tk in
// order to create all the appropriate widgets, and also 
// to write the matrix to the screen.  It is intended to
// be fully cross-platform.
//
// This class requires that there exist a global variable
// "interp," pointing to a valid Tcl Interpretter.
//
// This class requires a global int variable "cur_spyview"
// to contain a unique value each time a new view is
// constucted.  This is in order that the window may be
// assigned a unique name in tcl/tk.  It does not matter
// what this number is so long as it is never the same for
// any two instantiations, or else tcl/tk will break.  It
// is recommended that only numbers 000-999 be used, since
// only the last three digits will be added to the name.
//
// Note: This file is discluded from compilation if the
//       NO_TK_COMPILE option is specified in the compiler
//       flags (this is a convenience for folks without
//       the tcl/tk libraries)
//
// Author: Ben Langmead (btl10@columbia.edu)
// Date:   July 3, 2001
//
//////////////////////////////////////////////////////////

#ifndef NO_TK_COMPILE
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "../spmath.h"

//
// view::view()
//
//  Default constructor.  This is never used in the code.
// This simply initializes the internal variables and 
// creates a new viewer window without drawing a matrix.
//

view::view() {
  sp  = NULL;
  image_ht = 0;
  image_wd = 0;
  mat_xoff = 0;
  mat_yoff = 0;
  image_xoff = 0;
  image_yoff = 0;
  mat_ht = 0;
  mat_wd = 0;
  hphoto = NULL;
  hspread = NULL;
  dim = 0;
  gblock.pixelPtr  = (unsigned char *)NULL;
  gblock.width     = 0;
  gblock.height    = 0;
  gblock.pitch     = 0;
  gblock.pixelSize = 3;
  gblock.offset[0] = 0;
  gblock.offset[1] = 1;
  gblock.offset[2] = 2;
  memset(&stats, 0, sizeof(stats));
  memset(palette, 0, sizeof(view_rgb)*21);
  strcpy(targ, "");

  create_tk_window();
}

//
// view::view(sparse &spr)
//
//  Constructor.  Copies the matrix spr to an internal
// variable.  Initializes data, sets the matrix, gets
// matrix statistics with get_stats, opens a new
// toplevel window with create_tk_window, and finally
// draws the matrix onto the window's image space.
//

view::view(sparse &spr) {
  sp  = NULL;
  image_ht = 0;
  image_wd = 0;
  mat_xoff = 0;
  mat_yoff = 0;
  image_xoff = 0;
  image_yoff = 0;
  mat_ht = 0;
  mat_wd = 0;
  hphoto = NULL;
  hspread = NULL;
  dim = 0;
  gblock.pixelPtr  = (unsigned char *)NULL;
  gblock.width     = 0;
  gblock.height    = 0;
  gblock.pitch     = 0;
  gblock.pixelSize = 3;
  gblock.offset[0] = 0;
  gblock.offset[1] = 1;
  gblock.offset[2] = 2;
  memset(&stats, 0, sizeof(stats));
  memset(palette, 0, sizeof(view_rgb)*21);
  strcpy(targ, "");

  set_matrix(spr);
}

//
// view::view(dense &dens)
//
// Same as the constructor view::view(sparse *spr) except
// that set_matrix is given a dense argument
//

view::view(dense &dens) {
  sp  = NULL;
  image_ht = 0;
  image_wd = 0;
  mat_xoff = 0;
  mat_yoff = 0;
  image_xoff = 0;
  image_yoff = 0;
  mat_ht = 0;
  mat_wd = 0;
  hphoto = NULL;
  hspread = NULL;
  dim = 0;
  gblock.pixelPtr  = (unsigned char *)NULL;
  gblock.width     = 0;
  gblock.height    = 0;
  gblock.pitch     = 0;
  gblock.pixelSize = 3;
  gblock.offset[0] = 0;
  gblock.offset[1] = 1;
  gblock.offset[2] = 2;
  memset(&stats, 0, sizeof(stats));
  memset(palette, 0, sizeof(view_rgb)*21);
  strcpy(targ, "");

  set_matrix(dens);
}

//
// view::~view
//
//  Destructor.  Deletes the local copy of the matrix being
// displayed and calls delete_tk_window to close the viewer.
//

view::~view() {

  delete_tk_window();

  if(sp != NULL) {
    delete sp;
    sp = NULL;
  }
  if(gblock.pixelPtr != NULL) {
    delete[] gblock.pixelPtr;
    gblock.pixelPtr = NULL;
  }
}

//
// view::create()
//
// This method calls the actual tcl procedure that creates
// the tk view.  Then it calls the C++ method (see sparse.C)
// that computes the various statistics about the matrix.
// Then the photo size is set according to the image_wd and
// image_ht global variables in view.tcl
//

void view::create() {

  // All views are toplevel tk windows with names of the
  // form "spywinXXX" where XXX is some 3-digit 
  // zero-padded number.  cur_spyview is a global 
  // variable that is incremented everytime a new spyview 
  // is created.  This ensures the uniqueness of the 
  // three-digit number, but is also an ugly hack

  snprintf(targ, 13, ".spywin%03d", cur_spyview);

  create_tk_window();

  set_labels();

  image_wd = atoi(Tcl_GetVar(interp, strdup("image_wd"), TCL_GLOBAL_ONLY));
  image_ht = atoi(Tcl_GetVar(interp, strdup("image_ht"), TCL_GLOBAL_ONLY));

  if(hphoto != NULL) {
    Tk_PhotoSetSize(hphoto, image_wd + image_xoff, image_ht + image_yoff);
  }
}

//
// view::create_tk_window
//
//  A method for creating a new window in tk for displaying
// the matrix.  It calls the tk procedure "create_view" in
// view.tcl then retrives the handles to the images for
// drawing, then calls "init_view" which resizes the window
// and sets the color palette
//

void view::create_tk_window() {
  char buf[128];
  char *img_path;

  sprintf(buf, "create_view %s", targ);
  TCL(buf);

  img_path = Tcl_GetVar(interp, strdup("image_path"), TCL_GLOBAL_ONLY);
  sprintf(buf, "%s.%s", targ, img_path);

  // As of TCL 8.x, Tk_FindPhoto takes the TCL interpretter
  // as its first parameter

#if (TCL_MAJOR_VERSION > 7)
  hphoto = Tk_FindPhoto(interp, buf);
#else
  hphoto = Tk_FindPhoto(buf);
#endif

  img_path = Tcl_GetVar(interp, strdup("color_spread_path"), TCL_GLOBAL_ONLY);
  sprintf(buf, "%s.%s", targ, img_path);

  // As of TCL 8.x, Tk_FindPhoto takes the TCL interpretter
  // as its first parameter

#if (TCL_MAJOR_VERSION > 7)
  hspread = Tk_FindPhoto(interp, buf);
#else
  hspread = Tk_FindPhoto(buf);
#endif

  if(hphoto == NULL) {
    cerr << "Error: could not retrieve image from new tk window" << endl;
    return;
  }
  if(hspread == NULL) {
    cerr << "Error: could not retrieve spread image from tk" << endl;
    return;
  }

  Tk_PhotoGetSize(hphoto, &image_wd, &image_ht);
  Tk_PhotoGetSize(hspread, &spread_wd, &spread_ht);

  // Set the palette
  sprintf(buf, "init_view %s", targ);
  TCL(buf);
}

//
// view::delete_tk_window
//

void view::delete_tk_window() {
  char buf[64];

  sprintf(buf, "delete_view %s", targ);
  TCL(buf);
}

//
// view::set_matrix
//
//  Assigns the matrix for this particular window, to be
// stored locally for drawing.
//

void view::set_matrix(sparse &s) {
  if(sp != NULL) {
    delete sp;
    sp = NULL;
  }

  sp = new sparse(s);

  dim = sp->get_ncol();
  mat_ht = dim;
  mat_wd = dim;
}

//
// view::set_matrix
//
//  Assigns the matrix for this particular window, to be
// stored locally for drawing.  The dense matrix is first
// converted to a sparse one
//

void view::set_matrix(dense &d) {
  if(sp != NULL) {
    delete sp;
    sp = NULL;
  }

  sp = new sparse(d.get_nrow(), d.get_ncol(),
		  d.get_pointer());

  dim = sp->get_ncol();
  mat_ht = dim;
  mat_wd = dim;
}


//
// view::set_labels
//
//  Given the position of the mouse cursor (x, y) and
// the state of the left mouse button (ms), calculate
// the rectangle to be drawn on the image, and recalc
// the "X" "Y" and "Value" labels to reflect the element
// under the cursor.
//

void view::mouse_track(int x, int y, int ms) {
  char buf[128];

  assert(mat_wd != 0);
  assert(mat_ht != 0);

  if(image_wd == 0 || image_ht == 0) {
    return;
  }

  if(x < image_xoff || x >= (image_xoff+image_wd) ||
     y < image_yoff || y >= (image_yoff+image_ht)) {
    return;
  }

  x -= image_xoff;
  y -= image_yoff;

  static int last_el_x = -1;
  static int last_el_y = -1;

  static int last_ms = 0;
  static int is_dragging = 0;
  static int drag_x_base = 0;
  static int drag_y_base = 0;

  // Calculate which element of the matrix the cursor
  // is covering

  double scale_x = image_wd/(double)mat_wd;
  double scale_y = image_ht/(double)mat_wd;  

  int el_x = (int)(x/scale_x)+mat_xoff+1;
  int el_y = (int)(y/scale_y)+mat_yoff+1;
  double el_val;

  int rx1, ry1, rx2, ry2;

  // Check to see if the mouse button was just pressed
  // or released

  if(!last_ms && ms) {
    is_dragging = 1;
    drag_x_base = el_x;
    drag_y_base = el_y;
  }
  else if(last_ms && !ms) {
    int diff = abs(last_el_x - drag_x_base)+1;

    mat_wd = diff;
    mat_ht = diff;

    if(last_el_x < drag_x_base)
      mat_xoff = last_el_x -1;
    else
      mat_xoff = drag_x_base -1;

    if(last_el_y < drag_y_base)
      mat_yoff = last_el_y -1;
    else
      mat_yoff = drag_y_base -1;

    is_dragging = 0;
    last_ms = 0;

    status("Zooming in...");

    sprintf(buf, "clear_rectangles %s", targ);
    TCL(buf);

    update();

    sprintf(buf, "set_zoom_percentage %s {%.1f%%}",
	    targ, ((dim - mat_wd) / (float)(dim - 1))*100);
    TCL(buf);

    sprintf(buf, "set_vis_region %s %d %d",
	    targ, mat_wd, mat_ht);
    TCL(buf);

    draw();

    return;
  }

  last_ms = ms;

  // Short circuit if the cursor hasn't yet moved off
  // the square it's on

  if(el_x == last_el_x && el_y == last_el_y) {
    return;
  }


  // Now calculate the bounds of the rectangle

  // If we're in mid-drag, then recalculate the
  // bounds of the rectangle
  if(is_dragging) {
    int xdiff = abs(drag_x_base - el_x);
    int ydiff = abs(drag_y_base - el_y);

    if(xdiff > ydiff) {
      if(el_y < drag_y_base) {
	el_y = drag_y_base - xdiff;
	while(el_y <= mat_yoff) {
	  el_y ++;
	  (el_x < drag_x_base)? el_x++ : el_x--;
	}
      }
      else {
	el_y = drag_y_base + xdiff;
	while(el_y > mat_yoff+mat_ht) {
	  el_y --;
	  (el_x < drag_x_base)? el_x++ : el_x--;
	}
      }
    }
    else {
      if(el_x < drag_x_base) {
	el_x = drag_x_base - ydiff;
	while(el_x <= mat_xoff) {
	  el_x ++;
	  (el_y < drag_y_base)? el_y++ : el_y--;
	}
      }
      else {
	el_x = drag_x_base + ydiff;
	while(el_x > mat_xoff+mat_wd) {
	  el_x --;
	  (el_y < drag_y_base)? el_y++ : el_y--;
	}
      }
    }

    if(el_x < drag_x_base) {
      if(el_y < drag_y_base) {
	rx1 = (int)((el_x-mat_xoff-1)*scale_x);
	ry1 = (int)((el_y-mat_yoff-1)*scale_y);
	rx2 = (int)((drag_x_base-mat_xoff)*scale_x);
	ry2 = (int)((drag_y_base-mat_yoff)*scale_y);
      }
      else {
	rx1 = (int)((el_x-mat_xoff-1)*scale_x);
	ry1 = (int)((el_y-mat_yoff)*scale_y);
	rx2 = (int)((drag_x_base-mat_xoff)*scale_x);
	ry2 = (int)((drag_y_base-mat_yoff-1)*scale_y);
      }
    }
    else {
      if(el_y < drag_y_base) {
	rx1 = (int)((el_x-mat_xoff)*scale_x);
	ry1 = (int)((el_y-mat_yoff-1)*scale_y);
	rx2 = (int)((drag_x_base-mat_xoff-1)*scale_x);
	ry2 = (int)((drag_y_base-mat_yoff)*scale_y);
      }
      else {
	rx1 = (int)((el_x-mat_xoff)*scale_x);
	ry1 = (int)((el_y-mat_yoff)*scale_y);
	rx2 = (int)((drag_x_base-mat_xoff-1)*scale_x);
	ry2 = (int)((drag_y_base-mat_yoff-1)*scale_y);
      }
    }
  }
  // If we're not dragging, just calculate the
  // rectangle based on where the pointer is
  else {
    // Calulate one square's rectangle

    rx1 = (int)((el_x-mat_xoff-1)*scale_x);
    ry1 = (int)((el_y-mat_yoff-1)*scale_y);
    rx2 = (int)((el_x-mat_xoff)*scale_x);
    ry2 = (int)((el_y-mat_yoff)*scale_y);
  }

  last_el_x = el_x;
  last_el_y = el_y;

  // Now retrieve this element's value so we can
  // update the label

  if(sp != NULL) {
    el_val = sp->get_element(el_y, el_x);
  }
  else {
    el_val = -1.0;
  }

  // Update the labels
  
  sprintf(buf, "set_x %s %d", targ, el_x);
  TCL(buf);

  sprintf(buf, "set_y %s %d", targ, el_y);
  TCL(buf);

  sprintf(buf, "set_value %s {%g}", targ, el_val);
  TCL(buf);

  // Delete all previous rectangles drawn on the canvas

  sprintf(buf, "clear_rectangles %s", targ);
  TCL(buf);

  // Create the new rectangle as an object on the canvas

  sprintf(buf, "draw_rect %s %d %d %d %d", targ, 
	  rx1+image_xoff, ry1+image_yoff, 
	  rx2+image_xoff, ry2+image_yoff);
  TCL(buf);
}

void view::resize(int wd, int ht, int ix, int iy) {
  
  static int dont_redraw = 0;

  if(dont_redraw) { return; }

  if((wd == image_wd)   && (ht == image_ht) && 
     (ix == image_xoff) && (iy == image_yoff)) {

    return;
  }

  // Reset local variables to reflect the re-size

  image_wd = wd;
  image_ht = ht;
  image_xoff = ix;
  image_yoff = iy;

  // Save these in static vars for comparison later

  if(hphoto != NULL) {
    clear();
    status("Rendering Matrix...");

    // Ensures that resize will not be called again recursively
    dont_redraw = 1;
    update();
    dont_redraw = 0;

    Tk_PhotoSetSize(hphoto, wd + image_xoff, ht + image_yoff);

    draw();
  }
}

//
// view::zoom_out()
//
//  Zooms the viewer image out and, if necessary, redraws.
//

void view::zoom_out() {
  char buf[128];

  int zinc = dim/2;

  // Short circuit is already zoomed out all the way
  if(mat_wd == dim && mat_ht == dim) {
    return;
  }

  // Increate the height and width of the displayed submatrix
  mat_wd += zinc;
  if(mat_wd > dim) mat_wd = dim;
  mat_ht += zinc;
  if(mat_ht > dim) mat_ht = dim;

  // Decrease the x and y offsets of the displayed matrix
  mat_xoff -= zinc/2;
  if(mat_xoff < 0) mat_xoff = 0;
  mat_yoff -= zinc/2;
  if(mat_yoff < 0) mat_yoff = 0;

  // Make sure that we are not zoomed out further than all
  // the way
  while((mat_xoff + mat_wd) > dim) {
    mat_xoff --;
    if(mat_xoff < 0) {
      mat_xoff = 0;
      mat_wd --;
    }
  }
  while((mat_yoff + mat_ht) > dim) {
    mat_yoff --;
    if(mat_yoff < 0) {
      mat_yoff = 0;
      mat_ht--;
    }
  }

  // Remove the zoom rectangle
  sprintf(buf, "clear_rectangles %s", targ);
  TCL(buf);

  status("Zooming out...");

  update();

  sprintf(buf, "set_zoom_percentage %s {%.1f%%}",
	  targ, ((dim - mat_wd) / (float)(dim - 1))*100);
  TCL(buf);

  sprintf(buf, "set_vis_region %s %d %d",
	  targ, mat_wd, mat_ht);
  TCL(buf);

  // Redraw the new image
  draw();
}

//
// view::zoom(int percent)
//
// Given a percentage "percent," zoom the view to that
// percentage while remaining centered.  If the view
// cannot remain centered (becuase it would try to draw
// elements not within the bounds of the matrix) then
// adjust the offsets so that the zoom percentage is
// maintained but the centering is not.
//

void view::zoom(int percent) {
  char buf[128];

  int new_mat_wd = (int)((double)dim * (1.0-percent/100.0));
  int new_mat_ht = (int)((double)dim * (1.0-percent/100.0));

  if(new_mat_wd == mat_wd && new_mat_ht == mat_ht) {
    return;
  }
  if(new_mat_wd == 0) new_mat_wd = 1;
  if(new_mat_ht == 0) new_mat_ht = 1;

  if(new_mat_wd < mat_wd) {
    status("Zooming in...");
  } else {
    status("Zooming out...");
  }

  if(new_mat_wd > dim) {
    new_mat_wd = dim;
  }
  if(new_mat_ht > dim) {
    new_mat_ht = dim;
  }

  int new_mat_xoff = mat_xoff + (mat_wd - new_mat_wd)/2;
  int new_mat_yoff = mat_yoff + (mat_ht - new_mat_ht)/2;

  if(new_mat_xoff + new_mat_wd > dim) {
    new_mat_xoff -= new_mat_xoff + new_mat_wd - dim;
  }
  if(new_mat_xoff < 0) {
    new_mat_xoff = 0;
  }

  if(new_mat_yoff + new_mat_ht > dim) {
    new_mat_yoff -= new_mat_yoff + new_mat_ht - dim;
  }
  if(new_mat_yoff < 0) {
    new_mat_yoff = 0;
  }

  mat_wd = new_mat_wd;
  mat_ht = new_mat_ht;

  mat_xoff = new_mat_xoff;
  mat_yoff = new_mat_yoff;

  sprintf(buf, "clear_rectangles %s", targ);
  TCL(buf);

  update();

  sprintf(buf, "set_zoom_percentage %s {%.1f%%}",
	  targ, ((dim - mat_wd) / (float)(dim - 1))*100);
  TCL(buf);

  sprintf(buf, "set_vis_region %s %d %d",
	  targ, mat_wd, mat_ht);
  TCL(buf);

  // Redraw the new image
  draw();
}

//
// view::set_labels()
//
// Sets the labels in the main viewport to reflect the
// properties of the matrix being viewed.  Also fills the
// "more info" text box with all the appropriate stats.
//

void view::set_labels() {
  char buf[128];

  int ltitle;
  int lkey;

  char title_str[40];
  char key_str[9];
  char type_str[4];

  const char *title = sp->get_title();
  const char *type = sp->get_type();
  const char *key = sp->get_key();

  for(ltitle = 72; ltitle > 0; ltitle--) {
    if(title[ltitle-1] != ' ') break;
  }

  strncpy(title_str, title, 39);

  if(ltitle == 0) {
    strcpy(title_str, "(Untitled Matrix)");
  }
  else if(ltitle >= 37) {
    title_str[36] = '.';
    title_str[37] = '.';
    title_str[38] = '.';
    title_str[39] = '\0';
  } else {
    title_str[ltitle] = '\0';
  }

  for(lkey = 8; lkey > 0; lkey--) {
    if(key[lkey-1] != ' ') break;
  }

  strncpy(key_str, key, lkey);
  strncpy(type_str, type, 3);

  key_str[lkey] = '\0';
  type_str[3] = '\0';

  // Set the labels in the main window

  sprintf(buf, "set_title %s {%s}", 
	  targ, title_str);
  TCL(buf);

  //  sprintf(buf, "set_key %s {%s}",
  //	  targ, key_str);
  //  TCL(buf);

  //  sprintf(buf, "set_type %s {%s}", 
  //	  targ, type_str);
  //  TCL(buf);

  sprintf(buf, "set_dimension %s %d %d",
	  targ, dim, dim);
  TCL(buf);

  sprintf(buf, "set_nnz %s %d",
	  targ, sp->get_nnz());
  TCL(buf);

  sprintf(buf, "set_zoom_percentage %s {%.1f%%}",
	  targ, ((dim - mat_wd) / (float)(dim - 1))*100);
  TCL(buf);

  sprintf(buf, "set_vis_region %s %d %d",
	  targ, mat_wd, mat_ht);
  TCL(buf);

  double max = sp->max_element();

  sprintf(buf, "set_spread_limits %s {%.3f} {%.3f}", 
	  targ, -max, max);
  TCL(buf);
}

void view::set_stats() {
  char text[4096];
  char buf[128];

  int l, i;

  if(stats.Dimension > 0) { return; }

  int ltitle;
  int lkey;

  char title_str[40];
  char key_str[9];
  char type_str[4];

  const char *title = sp->get_title();
  const char *type = sp->get_type();
  const char *key = sp->get_key();

  for(ltitle = 72; ltitle > 0; ltitle--) {
    if(title[ltitle-1] != ' ') break;
  }

  strncpy(title_str, title, 39);

  if(ltitle == 0) {
    strcpy(title_str, "(Untitled Matrix)");
  }
  else if(ltitle >= 37) {
    title_str[36] = '.';
    title_str[37] = '.';
    title_str[38] = '.';
    title_str[39] = '\0';
  } else {
    title_str[ltitle] = '\0';
  }

  for(lkey = 8; lkey > 0; lkey--) {
    if(key[lkey-1] != ' ') break;
  }

  strncpy(key_str, key, lkey);
  strncpy(type_str, type, 3);

  key_str[lkey] = '\0';
  type_str[3] = '\0';

  status("Calculating Statistics...");
  update();

  sp->get_stats(stats);

  sprintf(text, "set_info_text %s {", targ);

  strcat(text, title);
  strcat(text, "\nKey  :  ");
  strcat(text, key_str);
  strcat(text, "\nType :  ");
  strcat(text, type_str);

  strcat(text, "\n\nDimension N"
	 "                                      =  ");
  sprintf(buf, "%d", stats.Dimension);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Number of nonzero elements"
	 "                       =  ");
  sprintf(buf, "%d", stats.NNz);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Average number of nonzero elements/Column"
	 "        =  ");
  sprintf(buf, "%.3E", stats.Average_NNZ_Per_Column);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Standard deviation for above average"
	 "             =  ");
  sprintf(buf, "%.3E", stats.Averate_NNZ_Std_Deviation);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Nonzero elements in strict lower part"
	 "            =  ");
  sprintf(buf, "%d", stats.NNZ_In_Strict_Lower_Part);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Nonzero elements in strict upper part"
	 "            =  ");
  sprintf(buf, "%d", stats.NNZ_In_Strict_Upper_Part);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Nonzero elements in main diagonal"
	 "                =  ");
  sprintf(buf, "%d", stats.NNZ_In_Main_Diagonal);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Weight of longest column"
	 "                         =  ");
  sprintf(buf, "%d", stats.Weight_Of_Longest_Column);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Weight of shortest column"
	 "                        =  ");
  sprintf(buf, "%d", stats.Weight_Of_Shortest_Column);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Weight of longest row"
	 "                            =  ");
  sprintf(buf, "%d", stats.Weight_Of_Longest_Row);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Weight of shortest row"
	 "                           =  ");
  sprintf(buf, "%d", stats.Weight_Of_Shortest_Row);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Matching elements in symmetry"
	 "                    =  ");
  sprintf(buf, "%d", stats.Matching_Elements_In_Symmetry);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Relative Symmetry Match (symmetry=1)"
	 "             =  ");
  sprintf(buf, "%.3E", stats.Relative_Symmetry_Match);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Average distance of a(i,j) from diag"
	 "             =  ");
  sprintf(buf, "%.3E", stats.Average_Distance_From_Diag);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Standard deviation for above average"
	 "             =  ");
  sprintf(buf, "%.3E", stats.Average_Distance_Std_Deviation);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");
  strcat(text, "-------------------------------------------------------------");
  strcat(text, "\n"); // End of section

  strcat(text, "Frobenius norm of A"
	 "                              =  ");
  sprintf(buf, "%.3E", stats.Frobenius_Norm);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Frobenius norm of symmetric part"
	 "                 =  ");
  sprintf(buf, "%.3E", stats.Frobenius_Norm_Of_Symmetric);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Frobenius norm of nonsymmetric part"
	 "              =  ");
  sprintf(buf, "%.3E", stats.Frobenius_Norm_Of_Nonsymmetric);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Maximum element in A"
	 "                             =  ");
  sprintf(buf, "%.3E", stats.Maximum_Element);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Percentage of weakly diagonally dominant rows"
	 "    =  ");
  sprintf(buf, "%.3E", stats.Percentage_Diag_Dominant_Rows);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Percentage of weakly diagonally dominant columns"
	 " =  ");
  sprintf(buf, "%.3E", stats.Percentage_Diag_Dominant_Columns);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");
  strcat(text, "-------------------------------------------------------------");
  strcat(text, "\n"); // End of section

  strcat(text, "Lower bandwidth  (max: i-j, a(i,j) .ne. 0)"
	 "       =  ");
  sprintf(buf, "%d", stats.Lower_Bandwidth);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Upper bandwidth  (max: j-i, a(i,j) .ne. 0)"
	 "       =  ");
  sprintf(buf, "%d", stats.Upper_Bandwidth);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Maximum Bandwidth"
	 "                                =  ");
  sprintf(buf, "%d", stats.Maximum_Bandwidth);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Average Bandwidth"
	 "                                =  ");
  sprintf(buf, "%.3E", stats.Average_Bandwidth);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "Number of nonzeros in skyline storage"
	 "            =  ");
  sprintf(buf, "%d", stats.NNZ_In_Skyline_Storage);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "90% of matrix is in the band of width"
	 "            =  ");
  sprintf(buf, "%d", stats.Ninety_Percent_Bandwidth);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "80% of matrix is in the band of width"
	 "            =  ");
  sprintf(buf, "%d", stats.Eighty_Percent_Bandwidth);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);
  strcat(text, "\n");

  strcat(text, "The total number of nonvoid diagonals is"
	 "         =  ");
  sprintf(buf, "%d", stats.N_Nonvoid_Diagonals);
  l = 9-strlen(buf);
  for(i = 0; i < l; i++) {
    strcat(text, " ");
  }
  strcat(text, buf);

  sprintf(buf, "\nThe %d most important diagonals are (offsets) :\n", 
	  stats.N_Important_Diagonals);
  strcat(text, buf);

  for(i = 0; i < stats.N_Important_Diagonals; i++) {
    char minibuf[8];
    int j;

    sprintf(minibuf, "%d", stats.Most_Important_Diagonals[i]);
    l = 4-strlen(minibuf);
    for(j = 0; j < l; j++) {
      strcat(text, " ");
    }
    strcat(text, minibuf);
    strcat(text, "  ");
  }

  strcat(text, "\nThe accumulated percentages they represent are   :\n");

  for(i = 0; i < stats.N_Important_Diagonals; i++) {
    int j;
    char minibuf[8];

    sprintf(minibuf, "%.1f", stats.Important_Diag_Accumulated_Percentages[i]);
    l = 4-strlen(minibuf);
    for(j = 0; j < l; j++) {
      strcat(text, " ");
    } 
    strcat(text, minibuf);
    strcat(text, "  ");
  }

  strcat(text, "\n");
  strcat(text, "-------------------------------------------------------------");
  strcat(text, "\n"); // End of section


  if(!stats.Has_Block_Structure) {
    strcat(text, "The matrix does not have a block structure\n");
  }
  else {
    strcat(text, "The matrix has a block structure\n");
  }

  strcat(text, "-------------------------------------------------------------");
  strcat(text, "\n"); // End of section

  strcat(text, "}");
  TCL(text);

  status("Done");
  update();
}

//
// view::recolor(unsigned char nr, unsigned char ng, unsigned char nb,
//               unsigned char pr, unsigned char pg, unsigned char pb)
//
// Takes the two rgb triples representing the greatest 
//
//

void view::recolor(unsigned char nr, unsigned char ng, unsigned char nb,
		   unsigned char pr, unsigned char pg, unsigned char pb) {

  int dr = (pr-nr)/20;
  int dg = (pg-ng)/20;
  int db = (pb-nb)/20;

  int i;

  Tk_PhotoGetSize(hspread, &spread_wd, &spread_ht);

  int bwd = spread_wd/19;

  for(i = 0; i < 20; i++) {
    unsigned char r = nr + i*dr;
    unsigned char g = ng + i*dg;
    unsigned char b = nb + i*db;

    palette[i].r = r;
    palette[i].g = g;
    palette[i].b = b;

    put_block(hspread, bwd*i, 0, bwd*(i+1), spread_ht, r, g, b);
  }

  palette[20].r = pr;
  palette[20].g = pg;
  palette[20].b = pb;
}

//
// view::draw()
//
// Checks a few preconditions and then calls 
// draw_sparse()
//

void view::draw() {

  if(hphoto == NULL) {
    //    cout << "Skipping draw, hphoto is NULL" << endl;
    return;
  }
  if(image_wd == 0 ||
     image_ht == 0) { 
    //    cout << "Skipping draw, image dimension is zero" << endl;
    return; 
  }

  if(sp != NULL) {
    draw_sparse();
  }
}

//
// view::draw_sparse()
//
//  Draws the sparse matrix sp to the image space,
// using mat_xoff and mat_yoff to determine the
// submatrix's x and y offsets into the overall matrix
// and mat_wd and mat_ht to determine its height and
// width.
//

void view::draw_sparse() {

  if(image_wd == 0 || image_ht == 0) {
    //    cout << "Image size zero, skipping draw_sparse" << endl;
    return;
  }

  int putpixel_call_count = 0;
  int avg_region_call_count = 0;

  // Calculate scale factors, i.e., how large should
  // a single matrix element appear in the image space

  double mat_scale = (mat_wd/(double)image_wd);

  double scale_x = 1/mat_scale;
  double scale_y = 1/mat_scale;

  // image_wd is a number which represents how many pixels
  // across the image is.  skip represents how

  int *ia, *ja;
  double *a;

  double tmax = sp->max_element();

  //   cout << "In draw_sparse.  image_ht = " << image_ht 
  //   << " image_wd = " << image_wd << " mat_ht = " << mat_ht 
  //   << " mat_wd = " << mat_wd << " xoff = " << mat_xoff 
  //   << " yoff = " << mat_yoff << endl;

  int i, j, i1, j1, i2, j2, sx, sy, dx, dy, klo, khi, k, pe;
  double avg, m;

  clear();

  // If mat_scale is less than 1, then the number of pixels
  // across the image on the screen excedes the number of
  // elements across the matrix.  In this case, it is more
  // efficient to map "forwards" from the non-zero entries
  // in the matrix to the pixels on the screen...
  if(mat_scale < 1) {

    ia = sp->get_ia();
    ja = sp->get_ja();
    a = sp->get_a();

    // Prepare the block size so that calls to 
    // put_block_prepared will be preset with a pixel block 
    // of sufficient size (not necessary, but improves speed).

    set_max_block_size((int)ceil(scale_x), (int)ceil(scale_y));

    // Iterate through all the rows that are within the
    // bounds of the submatrix to be displayed

    for(i = mat_yoff; i < (mat_yoff + mat_ht); i++) {

      sy = (int)floor((i-mat_yoff)*scale_y);
      dy = (int)ceil((i+1-mat_yoff)*scale_y);
    
      klo = ia[i]-1;
      khi = ia[i+1]-1;
    
      // Now iterate through all the columns in row i 
      // that contain nonzero entries

      for(k = klo; k < khi; k++) {
	
	// Let j equal the column index
	
	j = ja[k]-1; 
	
	// If j is within the bounds of the submatrix to
	// be displayed, and has a nonzero value, then draw 
	// the element (i, j).
	
	if(j >= mat_xoff && j < (mat_xoff + mat_wd)) {
	
	  // The following test is useful becuase CSC format
	  // will still sometimes store zero elements if they
	  // occur in the main diagonal.
	  
	  if(a[k] != 0.0) {
	    
	    sx = (int)floor((j-mat_xoff)*scale_x);
	    dx = (int)ceil((j+1-mat_xoff)*scale_x);
	    
	    // Let the coloring vary with the magnitude of
	    // the element's value relative to the overall 
	    // matrix's maximum value, tmax.
	    
	    //	    fgreen = fabs(a[k])/tmax;
	    //	    tg = (unsigned char)(255*(1.0 / (1.0-log(fgreen))));
	    
	    m = fabs(a[k]/tmax);
	    m = (10*(1.0 / (1.0-log(m))));

	    if(a[k] < 0.0) {
	      pe = 10 - (int)m;
	    } else {
	      pe = 10 + (int)m;
	    }
	    
	    // Draw the region

	    if((dx - sx) > 2) {
	      put_block_prepared(sx+image_xoff+1, 
				 sy+image_yoff+1, 
				 dx+image_xoff-1, 
				 dy+image_yoff-1, 
				 palette[pe].r, 
				 palette[pe].g,
				 palette[pe].b);
	    }
	    else {
	      put_block_prepared(sx+image_xoff, 
				 sy+image_yoff, 
				 dx+image_xoff, 
				 dy+image_yoff, 
				 palette[pe].r, 
				 palette[pe].g,
				 palette[pe].b);
	    }
	  }
	}
      }
    }
  }

  // Otherwise, if mat_scale is greater than or equal to 1,
  // then the number of elements across the matrix excedes
  // the number of pixels across the display.  In this case,
  // it is more efficient to "backwards" map from the pixels
  // on the screen to their corresponding regions in the
  // matrix...
  else {

    // Calculate the offsets of the sides of the view

    int left   = image_xoff;
    int right  = image_xoff + image_wd;
    int top    = image_yoff;
    int bottom = image_yoff + image_ht;

    // Loop from the topmost pixel in the column to the
    // bottommost

    for(i = top; i < bottom; i++) {

      // set i1 and i2 to be the top and bottom rows
      // in this pixel's corresponding region in the
      // matrix
      
      i1 = (int)floor(((i-top)   * mat_scale));
      i2 = (int)ceil(((i-top+1) * mat_scale));

      if(i2 > mat_wd)
	continue;
      
      // Loop from the leftmost pixel to the rightmost

      for(j = left; j < right; j++) {
	
	// set j1 and j2 to be the left and right 
	// columns in this pixel's corresponding 
	// region in the matrix

	j1 = (int)floor(((j-left)   * mat_scale));
	j2 = (int)ceil(((j-left+1) * mat_scale));

	if(j2 > mat_ht)
	  continue;

	//	cout << "i1=" << i1 << " i2=" << i2 
	//	     << " j1=" << j1 << " j2=" << j2 << endl;
      
	if((i2-i1) && (j2-j1)) {
	  // Call the method that averages the
	  // nonzero entries in the region

	  if(j == left ) {
	    avg = sp->region_avg2(i1+mat_yoff, 
				  j1+mat_xoff, 
				  i2+mat_yoff, 
				  j2+mat_xoff,
				  0);
	  }
	  else {
	    avg = sp->region_avg2(i1+mat_yoff, 
				  j1+mat_xoff, 
				  i2+mat_yoff, 
				  j2+mat_xoff,
				  1);
	  }
	  
	  avg_region_call_count++;
	}
	else {
	  continue;
	}

	if(avg == 0.0) {
	  continue;
	}

	m = fabs(avg/tmax);
	m = (10*(1.0 / (1.0-log(m))));

	if(avg < 0.0) {
	  pe = 10 - (int)m;
	} else {
	  pe = 10 + (int)m;
	}

	// Draw the pixel

	putpixel_call_count++;
	put_pixel(j, i, palette[pe].r, palette[pe].g, palette[pe].b);
      }
    }
  }

  status("Done");

  //  cout << "Finished draw_sparse" << endl;
  //  cout << "Putpixel called " << putpixel_call_count << " times" << endl;
  //  cout << "Avg_region called " << avg_region_call_count << " times" << endl;
}

void view::set_max_block_size(int wd, int ht) {

  if(gblock.pixelPtr) {
    delete[] gblock.pixelPtr;
  }

  // Add a little extra padding just in case

  int elements = (wd+1) * (ht+1);

  if(elements) {
    gblock.pixelPtr = new unsigned char[elements * 3];
  }
}

//
// view::put_block
//

void view::put_block_prepared(int x1, int y1, int x2, int y2, 
		     unsigned char tr, unsigned char tg, unsigned char tb) {
  int c;
  int dataWidth  = x2-x1;
  int dataHeight = y2-y1;

  int elements = dataWidth * dataHeight * 3;

  for (c = 0; c < elements; c += 3) {
    gblock.pixelPtr[c]   = tr;
    gblock.pixelPtr[c+1] = tg;
    gblock.pixelPtr[c+2] = tb;
  }

  gblock.width  = dataWidth;
  gblock.height = dataHeight;
  gblock.pitch  = dataWidth * 3;

  Tk_PhotoPutBlock(hphoto, &gblock, x1, y1, dataWidth, dataHeight);
}

void view::put_block(Tk_PhotoHandle h,
		     int x1, int y1, int x2, int y2, 
		     unsigned char tr, unsigned char tg, unsigned char tb) {
  Tk_PhotoImageBlock block;

  int c;
  int dataWidth  = x2-x1;
  int dataHeight = y2-y1;

  int elements = dataWidth * dataHeight * 3;
  unsigned char *ptr = new unsigned char[elements];

  for (c = 0; c < elements; c += 3) {
    ptr[c]   = tr;
    ptr[c+1] = tg;
    ptr[c+2] = tb;
  }

  block.pixelPtr  = ptr;
  block.width     = dataWidth;
  block.height    = dataHeight;
  block.pitch     = dataWidth * 3;
  block.pixelSize = 3;
  block.offset[0] = 0;
  block.offset[1] = 1;
  block.offset[2] = 2;

  Tk_PhotoPutBlock(h, &block, x1, y1, dataWidth, dataHeight);

  delete block.pixelPtr;
}

//
// view::put_pixel(int x, int y, 
//	           unsigned char tr, unsigned char tg, unsigned char tb)
//
// Draws one pixel onto this view's image at (x, y) with color
// RGB(tr, tg, tb)
//

void view::put_pixel(int x, int y, 
		     unsigned char tr, unsigned char tg, unsigned char tb) {
  // This buffer will contain all

  static unsigned char prgb[3] = {
    255, 255, 255
  };

  static Tk_PhotoImageBlock pblock = {
    prgb,
    1, 1, 3, 3,
    {0, 1, 2}
  };

  prgb[0] = tr;
  prgb[1] = tg;
  prgb[2] = tb;

  Tk_PhotoPutBlock(hphoto, &pblock, x, y, 1, 1);
}

//
// view::clear()
//
// Clears the image and draws the white background square
//

void view::clear() {
  char buf[128];
  sprintf(buf, "blank_image %s", targ);
  TCL(buf);

  put_block(hphoto, image_xoff, image_yoff, image_xoff + image_wd, image_yoff + image_ht, 255, 255, 255);

}

//
// view::update()
//
// Little helper method to request that tk re-draw the view
//

void view::update() {
  TCL("update");
}

//
// view::status()
//
// Helper method that re-assigns the text for the status
// label at the bottom of the view.
//

void view::status(const char *s) {
  char buf[128];

  sprintf(buf, "set_status %s {%s}", targ, s);
  TCL(buf);
}
#endif
